import numpy as np


class Turbine:
    """
    An individual wind turbine.

    The turbine is merely a defined by basic data
    entries and a choice of turbine models.

    Attributes
    ----------
    xy: array_like
        The turbine ground position, shape: (2,)
    models: list of str
        The turbine model names, as they appear
        in the model book
    index: int, optional
        The index in the wind farm
    name: str, optional
        The turbine name/label
    mstates_sel: list of numpy.ndarray, optional
        For each turbine model, the state selection
        boolean array with shape (n_states,)
    D: float, optional
        The rotor diameter. Overwrites turbine type
        settings if given
    H: float, optional
        The hub height. Overwrites turbine type
        settings if given

    :group: core

    """

    def __init__(
        self,
        xy,
        turbine_models=[],
        index=None,
        name=None,
        models_state_sel=None,
        D=None,
        H=None,
    ):
        """
        Constructor.

        Parameters
        ----------
        xy: array_like
            The turbine ground position, shape: (2,)
        turbine_models: list of str
            The turbine model names, as they appear
            in the model book
        index: int, optional
            The index in the wind farm
        name: str, optional
            The turbine name/label
        models_state_sel: list of numpy.ndarray, optional
            For each turbine model, the state selection
            boolean array with shape (n_states,)
        D: float, optional
            The rotor diameter. Overwrites turbine type
            settings if given
        H: float, optional
            The hub height. Overwrites turbine type
            settings if given

        """
        self.index = index
        self.name = name
        self.xy = np.array(xy)
        self.models = turbine_models
        self.D = D
        self.H = H

        self.mstates_sel = models_state_sel
        if self.mstates_sel is None:
            self.mstates_sel = [None] * len(self.models)

    def add_model(self, model, states_sel=None):
        """
        Add a turbine model to the list.

        Parameters
        ----------
        model: foxes.core.TurbineModel
            The model
        states_sel: numpy.ndarray of bool, optional
            The states selection for the model, shape: (n_states,)

        """
        self.models.append(model)
        self.mstates_sel.append(states_sel)

    def insert_model(self, index, model, states_sel=None):
        """
        Insert a turbine model into the list of models.

        Parameters
        ----------
        index: int
            The position in the model list
        model: foxes.core.TurbineModel
            The model
        states_sel: numpy.ndarray of bool, optional
            The states selection for the model, shape: (n_states,)

        """
        self.models.insert(index, model)
        self.mstates_sel.insert(index, states_sel)



import numpy as np

from foxes.core import TurbineModel
import foxes.variables as FV


class Thrust2Ct(TurbineModel):
    """
    Calculates ct from thrust force data.

    Attributes
    ----------
    thrust_var: str
        Name of the thrust variable
    WSCT: str
        The wind speed variable for ct lookup

    :group: models.turbine_models

    """

    def __init__(self, thrust_var=FV.T, var_ws_ct=FV.REWS2):
        """
        Constructor.

        Parameters
        ----------
        thrust_var: str
            Name of the thrust variable
        var_ws_ct: str
            The wind speed variable for ct lookup

        """
        super().__init__()
        self.thrust_var = thrust_var
        self.WSCT = var_ws_ct

    def __repr__(self):
        a = f"thrust_var={self.thrust_var}, var_ws_ct={self.WSCT}"
        return f"{type(self).__name__}({a})"

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return [FV.CT]

    def calculate(self, algo, mdata, fdata, st_sel):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        st_sel: slice or numpy.ndarray of bool
            The state-turbine selection,
            for shape: (n_states, n_turbines)

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        ct = fdata[FV.CT]

        T = fdata[self.thrust_var][st_sel]
        rho = fdata[FV.RHO][st_sel]
        A = np.pi * (fdata[FV.D][st_sel] / 2) ** 2
        ws = fdata[self.WSCT][st_sel]

        ct[st_sel] = 2 * T / (rho * A * ws**2)

        return {FV.CT: ct}



import numpy as np
from numbers import Number

import foxes.constants as FC
import foxes.variables as FV

""" Mapping from windio to foxes variables
:group: input.windio
"""
wio2foxes = {
    "time": FC.TIME,
    "x": FV.X,
    "y": FV.Y,
    "height": FV.H,
    "wind_turbine": FC.TURBINE,
    "wind_direction": FV.WD,
    "wind_speed": FV.WS,
    "probability": FV.WEIGHT,
    "sector_probability": "sector_probability",
    "turbulence_intensity": FV.TI,
    "LMO": FV.MOL,
    "z0": FV.Z0,
    "reference_height": FV.H,
}

""" Mapping from foxes to windio variables
:group: input.windio
"""
foxes2wio = {d: k for k, d in wio2foxes.items()}

def _read_nondimensional_coordinate(name, wio_data, coords):
    """read nondimensional coordinate
    :group: input.windio
    """
    if isinstance(wio_data, Number):
        coords[wio2foxes[name]] = wio_data
        return True
    return False

def _read_dimensional_coordinate(name, wio_data, coords):
    """read dimensional coordinate
    :group: input.windio
    """
    if isinstance(wio_data, list):
        wio_data = np.array(wio_data)
    if isinstance(wio_data, np.ndarray) and len(wio_data.shape) == 1:
        coords[wio2foxes[name]] = wio_data
        return True
    return False

def _read_multi_dimensional_coordinate(name, wio_data, coords):
    """Read multi dimensional coordinate
    :group: input.windio
    """
    return _read_nondimensional_coordinate(
        name, wio_data, coords
    ) or _read_dimensional_coordinate(name, wio_data, coords)

def _read_nondimensional_data(name, wio_data, fields, dims):
    """read nondimensional data
    :group: input.windio
    """
    if isinstance(wio_data, Number):
        v = wio2foxes[name]
        fields[v] = wio_data
        dims[v] = []
        return True
    return False

def _read_dimensional_data(name, wio_data, fields, dims):
    """read dimensional data
    :group: input.windio
    """
    if isinstance(wio_data, dict) and "data" in wio_data and "dims" in wio_data:
        d = wio_data["data"]
        v = wio2foxes[name]
        fields[v] = d if isinstance(d, np.ndarray) else np.array(d)
        dims[v] = tuple([wio2foxes[c] for c in wio_data["dims"]])
        if len(dims[v]) != len(fields[v].shape):
            raise ValueError(
                f"Field '{name}': Dimensions {dims[v]} do not match shape {fields[v].shape}"
            )
        return True
    return False

def _read_multi_dimensional_data(name, wio_data, fields, dims):
    """Read multi dimensional data
    :group: input.windio
    """
    return _read_nondimensional_data(
        name, wio_data, fields, dims
    ) or _read_dimensional_data(name, wio_data, fields, dims)

def read_wind_resource_field(
    name, 
    wio_data, 
    coords, 
    fields, 
    dims, 
    verbosity,
    ):
    """
    Reads wind resource data into fields and dims

    Parameters
    ----------
    name: str
        The windio variable name
    wio_data: object
        The windio data
    coords: dict
        The coordinates dict, filled on the fly
    fields: dict
        The fields dict, filled on the fly
    dims: dict
        The dimensions dict, filled on the fly
    verbosity: int
        The verbosity level, 0=silent

    Returns
    -------
    success: bool
        Flag for successful data extraction

    :group: input.windio

    """
    if name in [
        "weibull_a",
        "weibull_k",
        "potential_temperature",
        "friction_velocity",
        "k",
        "epsilon",
        "ABL_height",
        "lapse_rate",
        "capping_inversion_thickness",
        "capping_inversion_strength",
    ]:
        if verbosity > 2:
            print(f"        Ignoring variable '{name}'")
        return False

    if verbosity > 2:
        print(f"        Reading variable '{name}'")
    if name in ["time", "wind_turbine"] and _read_multi_dimensional_coordinate(
        name, wio_data, coords
    ):
        return True

    elif name in [
        "wind_direction",
        "wind_speed",
        "x",
        "y",
        "height",
    ] and (
        _read_multi_dimensional_coordinate(name, wio_data, coords)
        or _read_multi_dimensional_data(name, wio_data, fields, dims)
    ):
        return True

    elif name in [
        "probability",
        "sector_probability",
        "turbulence_intensity",
        "LMO",
        "z0",
        "reference_height",
    ] and _read_multi_dimensional_data(name, wio_data, fields, dims):
        return True

    else:
        raise NotImplementedError(f"No reading method implemented for field '{name}'")



import numpy as np

from foxes.core import TurbineModel, Data
import foxes.variables as FV
import foxes.constants as FC


class RotorCentreCalc(TurbineModel):
    """
    Calculates data at the rotor centre

    Attributes
    ----------
    calc_vars: dict
        The variables that are calculated by the model,
        keys: var names, values: rotor var names

    :group: models.turbine_models

    """

    def __init__(self, calc_vars):
        """
        Constructor.

        Parameters
        ----------
        calc_vars: dict
            The variables that are calculated by the model,
            keys: var names, values: rotor var names

        """
        super().__init__()

        if isinstance(calc_vars, dict):
            self.calc_vars = calc_vars
        else:
            self.calc_vars = {v: v for v in calc_vars}

    def initialize(self, algo, verbosity=0):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        self._wcalc = algo.get_model("PointWakesCalculation")()
        super().initialize(algo, verbosity)

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            Names of all sub models

        """
        return [self._wcalc]

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return list(self.calc_vars.keys())

    def calculate(self, algo, mdata, fdata, st_sel):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        st_sel: slice or numpy.ndarray of bool
            The state-turbine selection,
            for shape: (n_states, n_turbines)

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        # prepare target point data:
        tdata = Data.from_points(
            fdata[FV.TXYH],
            data={
                v: np.zeros_like(fdata[FV.X][:, :, None])
                for v in self.calc_vars.values()
            },
            dims={v: (FC.STATE, FC.TARGET, FC.TPOINT) for v in self.calc_vars.values()},
            name=f"{self.name}_tdata",
        )

        # run ambient calculation:
        res = algo.states.calculate(algo, mdata, fdata, tdata)
        for v, a in FV.var2amb.items():
            if v in res:
                res[a] = res[v].copy()
        tdata.update(res)

        # run wake calculation:
        res = self._wcalc.calculate(algo, mdata, fdata, tdata)

        # extract results:
        out = {v: fdata[v] for v in self.calc_vars.keys()}
        for v in out.keys():
            out[v][st_sel] = res[self.calc_vars[v]][st_sel]

        return out



import numpy as np

from foxes.core import WakeK
from foxes.models.wake_models.gaussian import GaussianWakeModel
import foxes.variables as FV
import foxes.constants as FC


class Bastankhah2014(GaussianWakeModel):
    """
    The Bastankhah 2014 wake model

    Notes
    -----
    Reference:
    "A new analytical model for wind-turbine wakes"
    Majid Bastankhah, Fernando Porté-Agel
    https://doi.org/10.1016/j.renene.2014.01.002

    Attributes
    ----------
    sbeta_factor: float
        Factor multiplying sbeta
    induction: foxes.core.AxialInductionModel or str
        The induction model
    wake_k: foxes.core.WakeK
        Handler for the wake growth parameter k

    :group: models.wake_models.wind

    """

    def __init__(self, superposition, sbeta_factor=0.2, induction="Madsen", **wake_k):
        """
        Constructor.

        Parameters
        ----------
        superposition: str
            The wind speed deficit superposition.
        sbeta_factor: float
            Factor multiplying sbeta
        induction: foxes.core.AxialInductionModel or str
            The induction model
        wake_k: dict, optional
            Parameters for the WakeK class

        """
        super().__init__(superpositions={FV.WS: superposition})

        self.sbeta_factor = sbeta_factor
        self.induction = induction
        self.wake_k = WakeK(**wake_k)

    def __repr__(self):
        iname = (
            self.induction if isinstance(self.induction, str) else self.induction.name
        )
        s = f"{type(self).__name__}"
        s += f"({self.superpositions[FV.WS]}, induction={iname}, "
        s += self.wake_k.repr() + ")"
        return s

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            All sub models

        """
        return [self.wake_k, self.induction]

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        if isinstance(self.induction, str):
            self.induction = algo.mbook.axial_induction[self.induction]
        super().initialize(algo, verbosity, force)

    def calc_amplitude_sigma(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        x,
    ):
        """
        Calculate the amplitude and the sigma,
        both depend only on x (not on r).

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The x values, shape: (n_states, n_targets)

        Returns
        -------
        amsi: tuple
            The amplitude and sigma, both numpy.ndarray
            with shape (n_st_sel,)
        st_sel: numpy.ndarray of bool
            The state-target selection, for which the wake
            is non-zero, shape: (n_states, n_targets)

        """
        # get ct:
        ct = self.get_data(
            FV.CT,
            FC.STATE_TARGET,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=True,
        )
        
        # select targets:
        st_sel = (x > 0) & (ct > 0)
        if np.any(st_sel):
            # apply selection:
            x = x[st_sel]
            ct = ct[st_sel]

            # get D:
            D = self.get_data(
                FV.D,
                FC.STATE_TARGET,
                lookup="w",
                algo=algo,
                fdata=fdata,
                tdata=tdata,
                downwind_index=downwind_index,
                upcast=True,
            )[st_sel]

            # get k:
            k = self.wake_k(
                FC.STATE_TARGET,
                algo=algo,
                fdata=fdata,
                tdata=tdata,
                downwind_index=downwind_index,
                upcast=True,
            )[st_sel]

            # calculate sigma:
            # beta = 0.5 * (1 + np.sqrt(1.0 - ct)) / np.sqrt(1.0 - ct)
            a = self.induction.ct2a(ct)
            beta = (1 - a) / (1 - 2 * a)
            sigma = k * x + self.sbeta_factor * np.sqrt(beta) * D
            del beta, a

            # calculate amplitude:
            ct_eff = ct / (8 * (sigma / D) ** 2)
            ampld = np.maximum(-2 * self.induction.ct2a(ct_eff), -1)

        # case no targets:
        else:
            st_sel = np.zeros_like(x, dtype=bool)
            n_sp = np.sum(st_sel)
            ampld = np.zeros(n_sp, dtype=FC.DTYPE)
            sigma = np.zeros(n_sp, dtype=FC.DTYPE)

        return {FV.WS: (ampld, sigma)}, st_sel



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from windrose import WindroseAxes

FIGSIZE_DEFAULT = (8, 8)
DPI_DEFAULT = 80


class TabWindroseAxes(WindroseAxes):
    """
    A derivate of the wind rose axes that runs
    on stochastic data (bins with weights) instead
    of time series data

    :group: utils

    """

    @staticmethod
    def from_ax(
        ax=None,
        fig=None,
        figsize=None,
        rect=None,
        *args,
        **kwargs,
    ):
        """
        Return a WindroseAxes object for the figure `fig`.
        """
        if ax is None:
            if fig is None:
                fig = plt.figure(
                    figsize=FIGSIZE_DEFAULT if figsize is None else figsize,
                    dpi=DPI_DEFAULT,
                    facecolor="w",
                    edgecolor="w",
                )
            if rect is None:
                rect = [0.1, 0.1, 0.8, 0.8]
            ax = TabWindroseAxes(fig, rect, *args, **kwargs)
            fig.add_axes(ax)
            return ax
        else:
            return ax

    def _init_plot(self, direction, var, **kwargs):

        # self.clear()
        kwargs.pop("zorder", None)
        weights = kwargs.pop("weights")

        # Init of the bins array if not set
        if "bin_min_var" in kwargs:
            bins = list(kwargs.pop("bin_min_var"))
            bins.append(max(kwargs.pop("bin_max_var")))
        else:
            bins = kwargs.pop("bins", None)
            if bins is None:
                bins = np.linspace(np.min(var), np.max(var), 6)
            if isinstance(bins, int):
                bins = np.linspace(np.min(var), np.max(var), bins)
            bins = np.asarray(bins).tolist()
        nbins = len(bins)
        bins.append(np.inf)

        # Sets the colors table based on the colormap or the "colors" argument
        colors = kwargs.pop("colors", None)
        cmap = kwargs.pop("cmap", None)
        if colors is not None:
            if isinstance(colors, str):
                colors = [colors] * nbins
            if isinstance(colors, (tuple, list)):
                if len(colors) != nbins:
                    raise ValueError("colors and bins must have same length")
        else:
            if cmap is None:
                cmap = plt.get_cmap()
            colors = self._colors(cmap, nbins)

        if "bin_min_dir" in kwargs:
            angles = 90 - np.sort(np.unique(direction))
            angles[angles > 180] -= 360
            angles *= np.pi / 180

            dir_min = kwargs.pop("bin_min_dir")
            dir_edges = np.mod(dir_min, 360.0).tolist()

            dir_bins = dir_min.copy().tolist()
            if dir_bins[0] < 0:
                dir_bins.append(360 + dir_bins[0])
                dir_bins[0] = 0
                dir_bins.append(360 + dir_bins[1])

            nsector = len(angles)

        else:
            nsector = kwargs.pop("nsector", None)
            if nsector is None:
                nsector = 16
            angles = np.arange(0, -2 * np.pi, -2 * np.pi / nsector) + np.pi / 2

            angle = 360.0 / nsector
            dir_bins = np.arange(-angle / 2, 360.0 + angle, angle, dtype=float)
            dir_edges = dir_bins.tolist()

            dir_edges.pop(-1)
            dir_edges[0] = dir_edges.pop(-1)
            dir_bins[0] = 0.0

        table = np.histogram2d(
            x=var, y=direction, bins=[bins, dir_bins], density=False, weights=weights
        )[0]
        table[:, 0] = table[:, 0] + table[:, -1]
        table = table[:, :-1]

        self._info["dir"], self._info["bins"], self._info["table"] = (
            dir_edges,
            bins,
            table,
        )
        return bins, nbins, nsector, colors, angles, kwargs

    def legend(self, loc="upper right", *args, **kwargs):
        return super().legend(loc, *args, **kwargs)


if __name__ == "__main__":

    import numpy as np
    import pandas as pd
    from foxes import StaticData, STATES

    sdata = StaticData()
    fpath = sdata.get_file_path(STATES, "wind_rose_bremen.csv")
    print(fpath)
    data = pd.read_csv(fpath, index_col=0)
    print(data)

    wd = data["wd"].to_numpy()
    ws = data["ws"].to_numpy()
    weights = data["weight"].to_numpy()

    ax = TabWindroseAxes.from_ax()
    # ax.contourf(wd, ws, weights=weights, bins=[0,3,8,13], cmap=plt.cm.Blues)
    # ax.contour(wd, ws, weights=weights, bins=[0,3,8,13], colors='black')
    ax.bar(
        wd, ws, weights=weights, bins=[0, 3, 5, 8, 10, 13, 16, 20], cmap=plt.cm.Blues
    )
    ax.set_legend()
    plt.show()



"""
Functions for adding turbines to the wind farm.
"""

from .row import add_row
from .grid import add_grid
from .from_json import add_from_json
from .from_csv import add_from_csv
from .from_file import add_from_file
from .from_df import add_from_df
from .from_random import add_random



import foxes.variables as FV
import foxes.constants as FC

round_defaults = {v: 4 for v in FV.__dict__.keys() if isinstance(v, str)}
round_defaults[FV.WD] = 3
round_defaults[FV.YAW] = 3
round_defaults[FV.TI] = 6
round_defaults[FV.RHO] = 6
round_defaults[FC.XYH] = 3
round_defaults.update({FV.var2amb[v]: round_defaults[v] for v in FV.var2amb.keys()})



"""
Atmospheric input states.
"""

from .single import SingleStateStates
from .scan_ws import ScanWS
from .states_table import StatesTable, Timeseries, TabStates
from .field_data_nc import FieldDataNC
from .multi_height import MultiHeightStates, MultiHeightTimeseries
from .multi_height import MultiHeightNCStates, MultiHeightNCTimeseries

from . import create



import numpy as np
from xarray import Dataset

import foxes.constants as FC
import foxes.variables as FV
from foxes.utils import write_nc

from .output import Output


class PointCalculator(Output):
    """
    Computes results at given points

    Attributes
    ----------
    algo: foxes.Algorithm
        The algorithm for point calculation
    farm_results: xarray.Dataset
        The farm results

    :group: output

    """

    def __init__(self, algo, farm_results, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        algo: foxes.Algorithm
            The algorithm for point calculation
        farm_results: xarray.Dataset
            The farm results
        kwargs: dict, optional
            Additional parameters for the base class

        """
        super().__init__(**kwargs)
        self.algo = algo
        self.farm_results = farm_results

    def calculate(
        self,
        points,
        *args,
        states_mean=False,
        weight_turbine=0,
        to_file=None,
        write_vars=None,
        write_pars={},
        **kwargs,
    ):
        """
        Calculate point results

        Parameters
        ----------
        points: numpy.ndarray
            The points, shape: (n_points, 3)
            or (n_states, n_points, 3)
        args: tuple, optional
            Additional arguments for algo.calc_points
        states_mean: bool
            Flag for taking the mean over states
        weight_turbine: int, optional
            Index of the turbine from which to take the weight
        to_file: str, optional
            Path to the output netCDF file
        write_vars: list of str
            The variables to be written to file, or None
            for all
        write_pars: dict, optional
            Additional parameters for write_nc
        kwargs: tuple, optional
            Additional arguments for algo.calc_points

        Returns
        -------
        point_results: xarray.Dataset
            The point results. The calculated variables have
            dimensions (state, point)

        """
        if points.shape[-1] == 3 and len(points.shape) == 3:
            pts = points
            p_has_s = True
        elif points.shape[-1] == 3 and len(points.shape) == 2:
            pts = np.zeros([self.algo.n_states] + list(points.shape), dtype=FC.DTYPE)
            pts[:] = points[None, :]
            p_has_s = False
        else:
            raise ValueError(
                f"Expecting point shape (n_states, n_points, 3) or (n_points, 3), got {points.shape}"
            )

        pres = self.algo.calc_points(self.farm_results, pts, *args, **kwargs)

        if states_mean:
            weights = self.farm_results[FV.WEIGHT].to_numpy()[:, weight_turbine]
            pres = Dataset(
                data_vars={
                    v: np.einsum("s,sp->p", weights, pres[v].to_numpy())
                    for v in pres.data_vars.keys()
                }
            )

        vrs = list(pres.data_vars.keys()) if write_vars is None else write_vars
        if to_file is not None:
            if states_mean:
                if p_has_s:
                    points = np.einsum("s,spd->pd", weights, points)
                dvars = {
                    "x": ((FC.POINT,), points[..., 0]),
                    "y": ((FC.POINT,), points[..., 1]),
                    "z": ((FC.POINT,), points[..., 2]),
                }
                dvars.update({v: ((FC.POINT,), pres[v].to_numpy()) for v in vrs})
                ds = Dataset(data_vars=dvars)
            else:
                if p_has_s:
                    dvars = {
                        "x": ((FC.STATE, FC.POINT), points[..., 0]),
                        "y": ((FC.STATE, FC.POINT), points[..., 1]),
                        "z": ((FC.STATE, FC.POINT), points[..., 2]),
                    }
                else:
                    dvars = {
                        "x": ((FC.POINT,), points[..., 0]),
                        "y": ((FC.POINT,), points[..., 1]),
                        "z": ((FC.POINT,), points[..., 2]),
                    }
                dvars.update(
                    {v: ((FC.STATE, FC.POINT), pres[v].to_numpy()) for v in vrs}
                )
                ds = Dataset(
                    coords={FC.STATE: pres[FC.STATE].to_numpy()}, data_vars=dvars
                )

            fpath = self.get_fpath(to_file)
            write_nc(ds, fpath, **write_pars)

        return pres



from foxes.core import FarmController


class BasicFarmController(FarmController):
    """
    The basic farm controller, running
    all turbine models in their order, without
    conditions.

    :group: models.farm_controllers

    """

    pass



import numpy as np
from copy import deepcopy

from foxes.core import FarmDataModel, TData


class FarmWakesCalculation(FarmDataModel):
    """
    This model calculates wakes effects on farm data.

    :group: algorithms.downwind.models

    """

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        ovars = algo.rotor_model.output_farm_vars(
            algo
        ) + algo.farm_controller.output_farm_vars(algo)
        return list(dict.fromkeys(ovars))

    def calculate(self, algo, mdata, fdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        # collect ambient rotor results and weights:
        rotor = algo.rotor_model
        weights = rotor.from_data_or_store(rotor.RWEIGHTS, algo, mdata)
        amb_res = rotor.from_data_or_store(rotor.AMBRES, algo, mdata)

        # generate all wake evaluation points
        # (n_states, n_order, n_rpoints)
        pwake2tdata = {}
        for wname, wmodel in algo.wake_models.items():
            pwake = algo.partial_wakes[wname]
            if pwake.name not in pwake2tdata:
                tpoints, tweights = pwake.get_wake_points(algo, mdata, fdata)
                pwake2tdata[pwake.name] = TData.from_tpoints(tpoints, tweights)

        def _get_wdata(tdatap, wdeltas, s):
            """Helper function for wake data extraction"""
            tdata = tdatap.get_slice(s, keep=True)
            wdelta = {v: d[s] for v, d in wdeltas.items()}
            return tdata, wdelta

        def _evaluate(
            gmodel, tdata, amb_res, weights, wake_res, wdeltas, oi, wmodel, pwake
        ):
            """Helper function for data evaluation at turbines"""
            wres = gmodel.finalize_farm_wakes(
                algo, mdata, fdata, tdata, amb_res, weights, wdeltas, wmodel, oi, pwake
            )

            hres = {v: d[:, oi, None] for v, d in wake_res.items()}
            for v, d in wres.items():
                if v in wake_res:
                    hres[v] += d[:, None]

            rotor.eval_rpoint_results(
                algo, mdata, fdata, hres, weights, downwind_index=oi
            )

            res = algo.farm_controller.calculate(
                algo, mdata, fdata, pre_rotor=False, downwind_index=oi
            )
            fdata.update(res)

        wake_res = deepcopy(amb_res)
        n_turbines = mdata.n_turbines
        run_up = None
        run_down = None
        for wname, wmodel in algo.wake_models.items():
            pwake = algo.partial_wakes[wname]
            gmodel = algo.ground_models[wname]
            tdatap = pwake2tdata[pwake.name]
            wdeltas = gmodel.new_farm_wake_deltas(
                algo, mdata, fdata, tdatap, wmodel, pwake
            )

            # downwind:
            if wmodel.affects_downwind:
                run_up = wname
                for oi in range(n_turbines):
                    if oi > 0:
                        _evaluate(
                            gmodel,
                            tdatap,
                            amb_res,
                            weights,
                            wake_res,
                            wdeltas,
                            oi,
                            wmodel,
                            pwake,
                        )

                    if oi < n_turbines - 1:
                        tdata, wdelta = _get_wdata(tdatap, wdeltas, np.s_[:, oi + 1 :])
                        gmodel.contribute_to_farm_wakes(
                            algo, mdata, fdata, tdata, oi, wdelta, wmodel, pwake
                        )

            # upwind:
            else:
                run_down = wname
                for oi in range(n_turbines - 1, -1, -1):
                    if oi < n_turbines - 1:
                        _evaluate(
                            gmodel,
                            tdatap,
                            amb_res,
                            weights,
                            wake_res,
                            wdeltas,
                            oi,
                            wmodel,
                            pwake,
                        )

                    if oi > 0:
                        tdata, wdelta = _get_wdata(tdatap, wdeltas, np.s_[:, :oi])
                        gmodel.contribute_to_farm_wakes(
                            algo, mdata, fdata, tdata, oi, wdelta, wmodel, pwake
                        )

            if run_up is not None and run_down is not None:
                raise KeyError(
                    f"Wake model '{run_up}' is an upwind model, wake model '{run_down}' is a downwind model: Require iterative algorithm"
                )

        return {v: fdata[v] for v in self.output_farm_vars(algo)}



from abc import abstractmethod

from foxes.utils import all_subclasses

from .farm_data_model import FarmDataModel


class TurbineModel(FarmDataModel):
    """
    Abstract base class for turbine models.

    Turbine models are FarmDataModels that run
    on a selection of turbines.

    :group: core

    """

    @abstractmethod
    def calculate(self, algo, mdata, fdata, st_sel):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        st_sel: slice or numpy.ndarray of bool
            The state-turbine selection,
            for shape: (n_states, n_turbines)

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        pass

    @classmethod
    def new(cls, tmodel_type, *args, **kwargs):
        """
        Run-time turbine model factory.

        Parameters
        ----------
        tmodel_type: str
            The selected derived class name
        args: tuple, optional
            Additional parameters for constructor
        kwargs: dict, optional
            Additional parameters for constructor

        """

        if tmodel_type is None:
            return None

        allc = all_subclasses(cls)
        found = tmodel_type in [scls.__name__ for scls in allc]

        if found:
            for scls in allc:
                if scls.__name__ == tmodel_type:
                    return scls(*args, **kwargs)

        else:
            estr = "Turbine model type '{}' is not defined, available types are \n {}".format(
                tmodel_type, sorted([i.__name__ for i in allc])
            )
            raise KeyError(estr)



"""
Wind farm layout optimization problems.
"""

from .farm_layout import FarmLayoutOptProblem
from .regular_layout import RegularLayoutOptProblem
from .reggrids_layout import RegGridsLayoutOptProblem

from . import geom_layouts



from .segregated import PartialSegregated
from foxes.models.rotor_models.grid import GridRotor


class PartialGrid(PartialSegregated):
    """
    Partial wakes on a grid rotor that may
    differ from the one in the algorithm.

    :group: models.partial_wakes

    """

    def __init__(self, *args, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        args: tuple, optional
            Parameters for GridRotor
        kwargs: dict, optional
            Parameters for GridRotor

        """
        super().__init__(GridRotor(*args, calc_vars=[], **kwargs))



from abc import abstractmethod
import numpy as np

from foxes.core import WakeModel


class DistSlicedWakeModel(WakeModel):
    """
    Abstract base class for wake models for which
    the x-denpendency can be separated from the
    yz-dependency.

    The multi-yz ability is used by the `PartialDistSlicedWake`
    partial wakes model.

    Attributes
    ----------
    superpositions: dict
        The superpositions. Key: variable name str,
        value: The wake superposition model name,
        will be looked up in model book
    superp: dict
        The superposition dict, key: variable name str,
        value: `foxes.core.WakeSuperposition`

    :group: models.wake_models

    """

    def __init__(self, superpositions):
        """
        Constructor.

        Parameters
        ----------
        superpositions: dict
            The superpositions. Key: variable name str,
            value: The wake superposition model name,
            will be looked up in model book

        """
        super().__init__()
        self.superpositions = superpositions

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            Names of all sub models

        """
        return list(self.superp.values())

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        self.superp = {
            v: algo.mbook.wake_superpositions[s] for v, s in self.superpositions.items()
        }
        super().initialize(algo, verbosity, force)

    @abstractmethod
    def calc_wakes_x_yz(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        x,
        yz,
    ):
        """
        Calculate wake deltas.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The x values, shape: (n_states, n_targets)
        yz: numpy.ndarray
            The yz values for each x value, shape:
            (n_states, n_targets, n_yz_per_target, 2)

        Returns
        -------
        wdeltas: dict
            The wake deltas. Key: variable name str,
            value: numpy.ndarray, shape: (n_st_sel, n_yz_per_target)
        st_sel: numpy.ndarray of bool
            The state-target selection, for which the wake
            is non-zero, shape: (n_states, n_targets)

        """
        pass

    def contribute(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        wake_coos,
        wake_deltas,
    ):
        """
        Modifies wake deltas at target points by
        contributions from the specified wake source turbines.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints, ...)

        """
        # rounding for safe x > 0 conditions
        x = np.round(wake_coos[:, :, 0, 0], 12)
        yz = wake_coos[..., 1:3]

        wdeltas, st_sel = self.calc_wakes_x_yz(
            algo, mdata, fdata, tdata, downwind_index, x, yz
        )

        for v, hdel in wdeltas.items():
            try:
                superp = self.superp[v]
            except KeyError:
                raise KeyError(
                    f"Model '{self.name}': Missing wake superposition entry for variable '{v}', found {sorted(list(self.superp.keys()))}"
                )

            wake_deltas[v] = superp.add_wake(
                algo,
                mdata,
                fdata,
                tdata,
                downwind_index,
                st_sel,
                v,
                wake_deltas[v],
                hdel,
            )

    def finalize_wake_deltas(
        self,
        algo,
        mdata,
        fdata,
        amb_results,
        wake_deltas,
    ):
        """
        Finalize the wake calculation.

        Modifies wake_deltas on the fly.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        amb_results: dict
            The ambient results, key: variable name str,
            values: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints)
        wake_deltas: dict
            The wake deltas object at the selected target
            turbines. Key: variable str, value: numpy.ndarray
            with shape (n_states, n_targets, n_tpoints)

        """
        for v, s in self.superp.items():
            wake_deltas[v] = s.calc_final_wake_delta(
                algo, mdata, fdata, v, amb_results[v], wake_deltas[v]
            )



import importlib.resources as resources

from pathlib import Path


class DataBook:
    """
    Container class for file paths, either directly
    given or as static data within a package.

    Parameters
    ----------
    data_book: DataBook, optional
        A data book to start from

    Attributes
    ----------
    dbase: dict
        The data base. Key: context str,
        value: dict (file name str to pathlib.Path)

    :group: utils

    """

    def __init__(self, data_book=None):
        """
        Constructor.

        Parameters
        ----------
        data_book: DataBook, optional
            A data book to start from

        """
        self.dbase = {}
        if data_book is not None:
            for c, d in data_book.items():
                self.dbase[c] = {}
                self.dbase[c].update(d)

    def add_data_package(self, context, package, file_sfx):
        """
        Add static files from a package location.

        Parameters
        ----------
        context: str
            The context
        package: str or package
            The package, must contain init file
        file_sfx: list of str
            File endings to include

        """
        if context not in self.dbase:
            self.dbase[context] = {}

        if isinstance(file_sfx, str):
            file_sfx = [file_sfx]

        try:
            contents = [
                r.name for r in resources.files(package).iterdir() if r.is_file()
            ]
        except AttributeError:
            contents = list(resources.contents(package))

        check_f = lambda f: any(
            [len(f) > len(s) and f[-len(s) :] == s for s in file_sfx]
        )
        contents = [f for f in contents if check_f(f)]

        try:
            for f in contents:
                with resources.as_file(resources.files(package).joinpath(f)) as path:
                    self.dbase[context][f] = path
        except AttributeError:
            for f in contents:
                with resources.path(package, f) as path:
                    self.dbase[context][f] = path

    def add_data_package_file(self, context, package, file_name):
        """
        Add a static file from a package location.

        Parameters
        ----------
        context: str
            The context
        package: str or package
            The package, must contain init.py` file
        file_mane: str
            The file name

        """
        if context not in self.dbase:
            self.dbase[context] = {}

        try:
            with resources.path(package, file_name) as path:
                self.dbase[context][file_name] = path
        except FileNotFoundError:
            raise FileNotFoundError(
                f"File '{file_name}' not found in package '{package}'"
            )

    def add_files(self, context, file_paths):
        """
        Add file paths

        Parameters
        ----------
        context: str
            The context
        file_paths: list of str
            The file paths

        """

        if context not in self.dbase:
            self.dbase[context] = {}

        for f in file_paths:
            path = Path(f)
            if not path.is_file():
                raise FileNotFoundError(
                    f"File '{path}' not found, cannot add to context '{context}'"
                )
            self.dbase[context][path.name] = path

    def add_file(self, context, file_path):
        """
        Add a file path

        Parameters
        ----------
        context: str
            The context
        file_path: str
            The file path

        """
        self.add_files(context, [file_path])

    def get_file_path(self, context, file_name, check_raw=True, errors=True):
        """
        Get path of a file

        Parameters
        ----------
        context: str
            The context
        file_name: str
            The file name
        check_raw: bool
            Check if `file_name` exists as given, and in
            that case return the path
        errors: bool
            Flag for raising KeyError, otherwise return None,
            if context of file_name not found

        Returns
        -------
        path: pathlib.Path
            The path

        """

        if check_raw:
            path = Path(file_name)
            if path.is_file():
                return path

        file_name = str(file_name)

        try:
            cdata = self.dbase[context]
        except KeyError:
            if not errors:
                return None
            raise KeyError(
                f"Context '{context}' not found in data book. Available: {sorted(list(self.dbase.keys()))}"
            )

        try:
            return cdata[file_name]
        except KeyError:
            if not errors:
                return None
            raise KeyError(
                f"File '{file_name}' not found in context '{context}'. Available: {sorted(list(cdata.keys()))}"
            )

    def toc(self, context):
        """
        Get list of contents

        Parameters
        ----------
        context: str
            The context

        Returns
        -------
        keys: list of str
            The data keys

        """
        return sorted(list(self.dbase[context].keys()))



import numpy as np

from foxes.core import RotorModel
from foxes.utils import wd2uv, uv2wd
import foxes.variables as FV
import foxes.constants as FC


class CentreRotor(RotorModel):
    """
    The centre rotor model.

    Evaluates states at a single point, located
    at the rotor centre.

    :group: models.rotor_models

    """

    def n_rotor_points(self):
        """
        The number of rotor points

        Returns
        -------
        n_rpoints: int
            The number of rotor points

        """
        return 1

    def design_points(self):
        """
        The rotor model design points.

        Design points are formulated in rotor plane
        (x,y,z)-coordinates in rotor frame, such that
        - (0,0,0) is the centre point,
        - (1,0,0) is the point radius * n_rotor_axis
        - (0,1,0) is the point radius * n_rotor_side
        - (0,0,1) is the point radius * n_rotor_up

        Returns
        -------
        dpoints: numpy.ndarray
            The design points, shape: (n_points, 3)

        """
        return np.array([[0.0, 0.0, 0.0]])

    def rotor_point_weights(self):
        """
        The weights of the rotor points

        Returns
        -------
        weights: numpy.ndarray
            The weights of the rotor points,
            add to one, shape: (n_rpoints,)

        """
        return np.array([1.0])

    def get_rotor_points(self, algo, mdata, fdata):
        """
        Calculates rotor points from design points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        points: numpy.ndarray
            The rotor points, shape:
            (n_states, n_turbines, n_rpoints, 3)

        """
        return fdata[FV.TXYH][:, :, None, :]

    def eval_rpoint_results(
        self,
        algo,
        mdata,
        fdata,
        rpoint_results,
        weights,
        downwind_index=None,
        copy_to_ambient=False,
    ):
        """
        Evaluate rotor point results.

        This function modifies `fdata`, either
        for all turbines or one turbine per state,
        depending on parameter `states_turbine`. In
        the latter case, the turbine dimension of the
        `rpoint_results` is expected to have size one.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        rpoint_results: dict
            The results at rotor points. Keys: variable str.
            Values: numpy.ndarray, shape if `states_turbine`
            is None: (n_states, n_turbines, n_rpoints).
            Else: (n_states, 1, n_rpoints)
        weights: numpy.ndarray
            The rotor point weights, shape: (n_rpoints,)
        downwind_index: int, optional
            The index in the downwind order
        copy_to_ambient: bool
            If `True`, the fdata results are copied to ambient
            variables after calculation

        """
        if len(weights) > 1:
            return super().eval_rpoint_results(
                algo, mdata, fdata, rpoint_results, weights, downwind_index
            )

        n_states = mdata.n_states
        n_turbines = algo.n_turbines

        uvp = None
        uv = None
        if (
            FV.WS in self.calc_vars
            or FV.WD in self.calc_vars
            or FV.YAW in self.calc_vars
            or FV.REWS in self.calc_vars
            or FV.REWS2 in self.calc_vars
            or FV.REWS3 in self.calc_vars
        ):
            wd = rpoint_results[FV.WD]
            ws = rpoint_results[FV.WS]
            uvp = wd2uv(wd, ws, axis=-1)
            uv = uvp[:, :, 0]

        wd = None
        vdone = []
        for v in self.calc_vars:
            if v not in fdata:
                fdata[v] = np.zeros((n_states, n_turbines), dtype=FC.DTYPE)

            if v == FV.WD or v == FV.YAW:
                if wd is None:
                    wd = uv2wd(uv, axis=-1)
                self._set_res(fdata, v, wd, downwind_index)
                vdone.append(v)

            elif v == FV.WS:
                self._set_res(fdata, v, ws[:, :, 0], downwind_index)
                del ws
                vdone.append(v)
        del uv, wd

        if (
            FV.REWS in self.calc_vars
            or FV.REWS2 in self.calc_vars
            or FV.REWS3 in self.calc_vars
        ):
            if downwind_index is None:
                yaw = fdata[FV.YAW]
            else:
                yaw = fdata[FV.YAW][:, downwind_index, None]
            nax = wd2uv(yaw, axis=-1)
            wsp = np.einsum("stpd,std->stp", uvp, nax)

            for v in self.calc_vars:
                if v == FV.REWS or v == FV.REWS2 or v == FV.REWS3:
                    rews = wsp[:, :, 0]
                    self._set_res(fdata, v, rews, downwind_index)
                    del rews
                    vdone.append(v)

            del wsp
        del uvp

        for v in self.calc_vars:
            if v not in vdone:
                res = rpoint_results[v][:, :, 0]
                self._set_res(fdata, v, res, downwind_index)
                del res
            if copy_to_ambient and v in FV.var2amb:
                fdata[FV.var2amb[v]] = fdata[v].copy()



from abc import abstractmethod

from .point_data_model import PointDataModel, PointDataModelList
from foxes.utils import all_subclasses
import foxes.variables as FV
import foxes.constants as FC


class States(PointDataModel):
    """
    Abstract base class for states.

    States describe ambient meteorological data,
    typically wind speed, wind direction, turbulence
    intensity and air density.

    :group: core

    """

    @abstractmethod
    def size(self):
        """
        The total number of states.

        Returns
        -------
        int:
            The total number of states

        """
        pass

    def index(self):
        """
        The index list

        Returns
        -------
        indices: array_like
            The index labels of states, or None for default integers

        """
        return list(range(self.size()))

    @abstractmethod
    def weights(self, algo):
        """
        The statistical weights of all states.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        weights: numpy.ndarray
            The weights, shape: (n_states, n_turbines)

        """
        pass

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        idata = super().load_data(algo, verbosity)

        sinds = self.index()
        if sinds is not None:
            idata["coords"][FC.STATE] = sinds

        weights = self.weights(algo)
        if len(weights.shape) != 2:
            raise ValueError(
                f"States '{self.name}': Wrong weights dimension, expecing ({FC.STATE}, {FC.TURBINE}), got shape {weights.shape}"
            )
        if weights.shape[1] != algo.n_turbines:
            raise ValueError(
                f"States '{self.name}': Wrong size of second axis dimension '{FC.TURBINE}': Expecting {algo.n_turbines}, got {weights.shape[1]}"
            )
        idata["data_vars"][FV.WEIGHT] = ((FC.STATE, FC.TURBINE), weights)

        return idata

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return [FV.WS, FV.WD, FV.TI, FV.RHO]

    def __add__(self, s):
        if isinstance(s, list):
            return ExtendedStates(self, s)
        elif isinstance(s, ExtendedStates):
            if s.states is not self:
                raise ValueError(
                    f"Cannot add extended states, since not based on same states"
                )
            return ExtendedStates(self, s.pmodels.models[1:])
        else:
            return ExtendedStates(self, [s])

    @classmethod
    def new(cls, states_type, *args, **kwargs):
        """
        Run-time states factory.

        Parameters
        ----------
        states_type: str
            The selected derived class name
        args: tuple, optional
            Additional parameters for constructor
        kwargs: dict, optional
            Additional parameters for constructor

        """

        if states_type is None:
            return None

        allc = all_subclasses(cls)
        found = states_type in [scls.__name__ for scls in allc]

        if found:
            for scls in allc:
                if scls.__name__ == states_type:
                    return scls(*args, **kwargs)
        else:
            estr = "States type '{}' is not defined, available types are \n {}".format(
                states_type, sorted([i.__name__ for i in allc])
            )
            raise KeyError(estr)


class ExtendedStates(States):
    """
    States extended by point data models.

    Attributes
    ----------
    states: foxes.core.States
        The base states to start from
    pmodels: foxes.core.PointDataModelList
        The point models, including states as first model

    :group: core

    """

    def __init__(self, states, point_models=[]):
        """
        Constructor.

        Parameters
        ----------
        states: foxes.core.States
            The base states to start from
        point_models: list of foxes.core.PointDataModel, optional
            The point models, executed after states

        """
        super().__init__()
        self.states = states
        self.pmodels = PointDataModelList(models=[states] + point_models)

    def append(self, model):
        """
        Add a model to the list

        Parameters
        ----------
        model: foxes.core.PointDataModel
            The model to add

        """
        self.pmodels.append(model)

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            Names of all sub models

        """
        return [self.pmodels]

    def size(self):
        """
        The total number of states.

        Returns
        -------
        int:
            The total number of states

        """
        return self.states.size()

    def index(self):
        """
        The index list

        Returns
        -------
        indices: array_like
            The index labels of states, or None for default integers

        """
        return self.states.index()

    def weights(self, algo):
        """
        The statistical weights of all states.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        weights: numpy.ndarray
            The weights, shape: (n_states, n_turbines)

        """
        return self.states.weights(algo)

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.states.output_point_vars(algo)

    def calculate(self, algo, mdata, fdata, tdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_points)

        """
        return self.pmodels.calculate(algo, mdata, fdata, tdata)

    def __add__(self, m):
        models = self.pmodels.models[1:]
        if isinstance(m, list):
            return ExtendedStates(self.states, models + m)
        elif isinstance(m, ExtendedStates):
            if m.states is not self.states:
                raise ValueError(
                    f"Cannot add extended states, since not based on same states"
                )
            return ExtendedStates(self.states, models + m.pmodels.models[1:])
        else:
            return ExtendedStates(self.states, models + [m])



import numpy as np
import pandas as pd
from pathlib import Path

from foxes.core import WindFarm
from foxes.models import ModelBook
from foxes.utils import import_module, Dict
from foxes.data import StaticData, WINDIO

from .read_fields import read_wind_resource_field
from .get_states import get_states
from .read_farm import read_layout, read_turbine_type
from .read_attributes import read_attributes
from .runner import WindioRunner


def _read_site(wio, algo_dict, verbosity):
    """Reads the site information"""
    wio_site = Dict(wio["site"], name="site")
    if verbosity > 1:
        print("Reading site")
        print("  Name:", wio_site.pop("name", None))
        print("  Contents:", [k for k in wio_site.keys()])

    # ignore boundaries:
    if verbosity > 2:
        print("  Ignoring boundaries")

    # read energy_resource:
    energy_resource = Dict(wio_site["energy_resource"], name="energy_resource")
    if verbosity > 2:
        print("  Reading energy_resource")
        print("    Name:", energy_resource.pop("name", None))
        print("    Contents:", [k for k in energy_resource.keys()])

    # read wind_resource:
    wind_resource = Dict(energy_resource["wind_resource"], name="wind_resource")
    if verbosity > 2:
        print("    Reading wind_resource")
        print("      Name:", wind_resource.pop("name", None))
        print("      Contents:", [k for k in wind_resource.keys()])

    # read fields
    coords = Dict(name="coords")
    fields = Dict(name="fields")
    dims = Dict(name="dims")
    for n, d in wind_resource.items():
        read_wind_resource_field(n, d, coords, fields, dims, verbosity)
    if verbosity > 2:
        print("      Coords:")
        for c, d in coords.items():
            print(f"        {c}: Shape {d.shape}")
        print("      Fields:")
        for f, d in dims.items():
            if len(d):
                print(f"        {f}: Dims {d}, shape {fields[f].shape}")
            else:
                print(f"        {f} = {fields[f]}")

    algo_dict["states"] = get_states(coords, fields, dims, verbosity)


def _read_farm(wio, algo_dict, verbosity):
    """Reads the wind farm information"""
    wio_farm = Dict(wio["wind_farm"], name="wind_farm")
    if verbosity > 1:
        print("Reading wind farm")
        print("  Name:", wio_farm.pop("name", None))
        print("  Contents:", [k for k in wio_farm.keys()])

    # find REWS exponents:
    try:
        rotor_averaging = wio["attributes"]["analysis"]["rotor_averaging"]
        ws_exp_P = rotor_averaging["wind_speed_exponent_for_power"]
        ws_exp_ct = rotor_averaging["wind_speed_exponent_for_ct"]
    except KeyError:
        ws_exp_P = 1
        ws_exp_ct = 1

    # read turbine type:
    turbines = Dict(wio_farm["turbines"], name="turbines")
    ttype = read_turbine_type(turbines, algo_dict, ws_exp_P, ws_exp_ct, verbosity)

    # read layouts:
    layouts = Dict(wio_farm["layouts"], name="layouts")
    if verbosity > 2:
        print("    Reading layouts")
        print("      Contents:", [k for k in layouts.keys()])
    for lname, ldict in layouts.items():
        read_layout(lname, ldict, algo_dict, ttype, verbosity)


def read_windio(windio_yaml, verbosity=1):
    """
    Reads a complete WindIO case.

    This is the main entry point for windio case
    calculations.

    Parameters
    ----------
    windio_yaml: str
        Path to the windio yaml file
    verbosity: int
        The verbosity level, 0 = silent

    Returns
    -------
    runner: foxes.input.windio.WindioRunner
        The windio runner, call its run function
        for the complete exection

    :group: input.windio

    """

    wio_file = Path(windio_yaml)
    if not wio_file.is_file():
        wio_file = StaticData().get_file_path(WINDIO, wio_file, check_raw=False)

    if verbosity > 0:
        print(f"Reading windio file {wio_file}")

    yml_utils = import_module("windIO.utils.yml_utils", hint="pip install windio")
    wio = yml_utils.load_yaml(wio_file)

    if verbosity > 1:
        print("  Name:", wio.pop("name", None))
        print("  Contents:", [k for k in wio.keys()])

    algo_dict = Dict(
        algo_type="Downwind",
        mbook=ModelBook(),
        farm=WindFarm(),
        wake_models=[],
        verbosity=verbosity-3,
    )

    _read_site(wio, algo_dict, verbosity)
    _read_farm(wio, algo_dict, verbosity)
    
    out_dicts, odir = read_attributes(
        wio,
        algo_dict,
        verbosity,
    )

    if verbosity > 1:
        print("Creating windio runner")
    runner = WindioRunner(
                algo_dict, 
                output_dir=odir,
                output_dicts=out_dicts, 
                wio_input_data=wio, 
                verbosity=verbosity
            )

    return runner


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-f",
        "--file",
        help="The windio yaml file",
        default="windio_5turbines_timeseries.yaml",
    )
    args = parser.parse_args()

    runner = read_windio(args.file)

    runner.run()



import numpy as np

from foxes.core import PointDataModel
import foxes.variables as FV


class TKE2TI(PointDataModel):
    """
    Calculates TI from TKE, using TI = sqrt( 3/2 * TKE) / WS

    :group: models.point_models

    """

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return [FV.TI]

    def calculate(self, algo, mdata, fdata, pdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_points)

        """
        tke = pdata[FV.TKE]
        ws = pdata[FV.WS]

        return {FV.TI: np.sqrt(1.5 * tke) / ws}



import numpy as np
import pandas as pd
from xarray import Dataset, open_dataset
from pathlib import Path
from scipy.interpolate import interp1d

from foxes.core import States
from foxes.utils import PandasFileHelper
from foxes.data import STATES
import foxes.variables as FV
import foxes.constants as FC
from foxes.utils import wd2uv, uv2wd


class MultiHeightStates(States):
    """
    States with multiple heights data per entry.

    The input data is taken from a csv file or
    pandas data frame with columns. The format
    of the data columns is as in the following
    example for wind speed at heights 50, 60, 100 m:

    WS-50, WS-60, WS-100, ...

    Attributes
    ----------
    data_source: str or pandas.DataFrame
        Either path to a file or data
    ovars: list of str
        The output variables
    heights: list of float
        The heights at which to search data
    var2col: dict, optional
        Mapping from variable names to data column names
    fixed_vars: dict, optional
        Fixed uniform variable values, instead of
        reading from data
    pd_read_pars: dict, optional
        pandas file reading parameters
    states_sel: slice or range or list of int
        States subset selection
    states_loc: list
        State index selection via pandas loc function
    RDICT: dict
        Default pandas file reading parameters

    :group: input.states

    """

    RDICT = {"index_col": 0}

    def __init__(
        self,
        data_source,
        output_vars,
        heights,
        var2col={},
        fixed_vars={},
        pd_read_pars={},
        states_sel=None,
        states_loc=None,
        ipars={},
    ):
        """
        Constructor.

        Parameters
        ----------
        data_source: str or pandas.DataFrame
            Either path to a file or data
        output_vars: list of str
            The output variables
        heights: list of float
            The heights at which to search data
        var2col: dict, optional
            Mapping from variable names to data column names
        fixed_vars: dict, optional
            Fixed uniform variable values, instead of
            reading from data
        pd_read_pars: dict, optional
            pandas file reading parameters
        states_sel: slice or range or list of int, optional
            States subset selection
        states_loc: list, optional
            State index selection via pandas loc function
        ipars: dict, optional
            Parameters for scipy.interpolate.interp1d

        """
        super().__init__()

        self.data_source = data_source
        self.ovars = output_vars
        self.heights = np.array(heights, dtype=FC.DTYPE)
        self.rpars = pd_read_pars
        self.var2col = var2col
        self.fixed_vars = fixed_vars
        self.ipars = ipars
        self.states_sel = states_sel
        self.states_loc = states_loc

        self._solo = None
        self._weights = None
        self._N = None

    def reset(self, algo=None, states_sel=None, states_loc=None, verbosity=0):
        """
        Reset the states, optionally select states

        Parameters
        ----------
        states_sel: slice or range or list of int, optional
            States subset selection
        states_loc: list, optional
            State index selection via pandas loc function
        verbosity: int
            The verbosity level, 0 = silent

        """
        if self.initialized:
            if algo is None:
                raise KeyError(f"{self.name}: Missing algo for reset")
            elif algo.states is not self:
                raise ValueError(f"{self.states}: algo.states differs from self")
            self.finalize(algo, verbosity)
        self.states_sel = states_sel
        self.states_loc = states_loc

    def _find_cols(self, v, cols):
        """
        Helper function for searching height columns
        """
        c0 = self.var2col.get(v, v)
        if v in self.fixed_vars:
            return []
        elif c0 in cols:
            return [c0]
        else:
            cls = []
            for h in self.heights:
                hh = int(h) if int(h) == h else h
                c = f"{c0}-{hh}"
                oc = self.var2col.get(c, c)
                if oc in cols:
                    cls.append(oc)
                else:
                    raise KeyError(
                        f"Missing: '{v}' in fixed_vars, or '{c0}' or '{oc}' in columns. Maybe make use of var2col?"
                    )
            return cls

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        if not isinstance(self.data_source, pd.DataFrame):
            if not Path(self.data_source).is_file():
                if verbosity:
                    print(
                        f"States '{self.name}': Reading static data '{self.data_source}' from context '{STATES}'"
                    )
                self.data_source = algo.dbook.get_file_path(
                    STATES, self.data_source, check_raw=False
                )
                if verbosity:
                    print(f"Path: {self.data_source}")
            elif verbosity:
                print(f"States '{self.name}': Reading file {self.data_source}")
            rpars = dict(self.RDICT, **self.rpars)
            data = PandasFileHelper().read_file(self.data_source, **rpars)
            isorg = False
        else:
            isorg = True
            data = self.data_source

        if self.states_sel is not None:
            data = data.iloc[self.states_sel]
        elif self.states_loc is not None:
            data = data.loc[self.states_loc]

        self._N = len(data.index)
        self._inds = data.index.to_numpy()

        col_w = self.var2col.get(FV.WEIGHT, FV.WEIGHT)
        self._weights = np.zeros((self._N, algo.n_turbines), dtype=FC.DTYPE)
        if col_w in data:
            self._weights[:] = data[col_w].to_numpy()[:, None]
        elif FV.WEIGHT in self.var2col:
            raise KeyError(
                f"Weight variable '{col_w}' defined in var2col, but not found in states table columns {data.columns}"
            )
        else:
            self._weights[:] = 1.0 / self._N
            if isorg:
                data = data.copy()
            data[col_w] = self._weights[:, 0]

        cols = []
        cmap = {}
        self._solo = {}
        for v in self.ovars:
            vcols = self._find_cols(v, data.columns)
            if len(vcols) == 1:
                self._solo[v] = data[vcols[0]].to_numpy()
            elif len(vcols) > 1:
                cmap[v] = (len(cols), len(cols) + len(vcols))
                cols += vcols
        data = data[cols]

        self.H = self.var(FV.H)
        self.VARS = self.var("vars")
        self.DATA = self.var("data")

        idata = super().load_data(algo, verbosity)

        idata["coords"][self.H] = self.heights
        idata["coords"][self.VARS] = list(cmap.keys())

        n_hts = len(self.heights)
        n_vrs = int(len(data.columns) / n_hts)
        dims = (FC.STATE, self.VARS, self.H)
        idata["data_vars"][self.DATA] = (
            dims,
            data.to_numpy().reshape(self._N, n_vrs, n_hts),
        )

        for v, d in self._solo.items():
            idata["data_vars"][self.var(v)] = ((FC.STATE,), d)
        self._solo = list(self._solo.keys())

        return idata

    def size(self):
        """
        The total number of states.

        Returns
        -------
        int:
            The total number of states

        """
        return self._N

    def index(self):
        """
        The index list

        Returns
        -------
        indices: array_like
            The index labels of states, or None for default integers

        """
        return self._inds

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.ovars

    def weights(self, algo):
        """
        The statistical weights of all states.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        weights: numpy.ndarray
            The weights, shape: (n_states, n_turbines)

        """
        return self._weights

    def calculate(self, algo, mdata, fdata, tdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints)

        """
        n_states = tdata.n_states
        n_targets = tdata.n_targets
        n_tpoints = tdata.n_tpoints
        h = mdata[self.H]
        z = tdata[FC.TARGETS][..., 2].reshape(n_states, n_targets * n_tpoints)
        n_h = len(h)
        vrs = list(mdata[self.VARS])
        n_vars = len(vrs)

        coeffs = np.zeros((n_h, n_h), dtype=FC.DTYPE)
        np.fill_diagonal(coeffs, 1.0)
        ipars = dict(assume_sorted=True, bounds_error=True)
        ipars.update(self.ipars)
        intp = interp1d(h, coeffs, axis=0, **ipars)
        ires = intp(z)
        del coeffs, intp

        has_wd = FV.WD in vrs
        if has_wd:
            i_wd = vrs.index(FV.WD)
            if FV.WS in vrs:
                i_ws = vrs.index(FV.WS)
                uvh = wd2uv(
                    mdata[self.DATA][:, i_wd], mdata[self.DATA][:, i_ws], axis=-1
                )
            elif FV.WS in self.fixed_vars:
                uvh = wd2uv(mdata[self.DATA][:, i_wd], self.fixed_vars[FV.WS], axis=-1)
            elif self.var(FV.WS) in mdata:
                uvh = wd2uv(
                    mdata[self.DATA][:, i_wd], mdata[self.var(FV.WS)][:, None], axis=-1
                )
            else:
                raise KeyError(
                    f"States '{self.name}': Found variable '{FV.WD}', but missing variable '{FV.WS}'"
                )
            uv = np.einsum("shd,sph->spd", uvh, ires).reshape(
                n_states, n_targets, n_tpoints, 2
            )
            del uvh

        ires = np.einsum("svh,sph->vsp", mdata[self.DATA], ires).reshape(
            n_vars, n_states, n_targets, n_tpoints
        )

        results = {}
        for v in self.ovars:
            if has_wd and v == FV.WD:
                results[v] = uv2wd(uv, axis=-1)
            elif has_wd and v == FV.WS:
                results[v] = np.linalg.norm(uv, axis=-1)
            elif v in self.fixed_vars:
                results[v] = np.zeros((n_states, n_targets, n_tpoints), dtype=FC.DTYPE)
                results[v][:] = self.fixed_vars[v]
            elif v in self._solo:
                results[v] = np.zeros((n_states, n_targets, n_tpoints), dtype=FC.DTYPE)
                results[v][:] = mdata[self.var(v)][:, None, None]
            else:
                results[v] = ires[vrs.index(v)]

        return results

    def finalize(self, algo, verbosity=0):
        """
        Finalizes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level

        """
        super().finalize(algo, verbosity)
        self._solo = None
        self._weights = None
        self._N = None


class MultiHeightNCStates(MultiHeightStates):
    """
    Multi-height states from xarray Dataset.

    Attributes
    ----------
    data_source: str or xarray.Dataset
        Either path to a file or data
    state_coord: str
        Name of the state coordinate
    h_coord: str
        Name of the height coordinate
    xr_read_pars: dict
        Parameters for xarray.open_dataset

    :group: input.states

    """

    def __init__(
        self,
        data_source,
        *args,
        state_coord=FC.STATE,
        h_coord=FV.H,
        heights=None,
        format_times_func="default",
        xr_read_pars={},
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        data_source: str or pandas.DataFrame
            Either path to a file or data
        args: tuple, optional
            Parameters for the base class
        state_coord: str
            Name of the state coordinate
        h_coord: str
            Name of the height coordinate
        output_vars: list of str
            The output variables
        heights: list of float, optional
            The heights at which to search data
        format_times_func: Function or 'default', optional
            The function that maps state_coord values
            to datetime dtype format
        xr_read_pars: dict, optional
            Parameters for xarray.open_dataset
        kwargs: dict, optional
            Parameters for the base class

        """
        super().__init__(
            data_source,
            *args,
            heights=[],
            pd_read_pars=None,
            **kwargs,
        )
        self.state_coord = state_coord
        self.heights = heights
        self.h_coord = h_coord
        self.xr_read_pars = xr_read_pars

        if format_times_func == "default":
            self.format_times_func = lambda t: t.astype("datetime64[ns]")
        else:
            self.format_times_func = format_times_func

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        if not isinstance(self.data_source, Dataset):
            if not Path(self.data_source).is_file():
                if verbosity:
                    print(
                        f"States '{self.name}': Reading static data '{self.data_source}' from context '{STATES}'"
                    )
                self.data_source = algo.dbook.get_file_path(
                    STATES, self.data_source, check_raw=False
                )
                if verbosity:
                    print(f"Path: {self.data_source}")
            elif verbosity:
                print(f"States '{self.name}': Reading file {self.data_source}")
            data = open_dataset(self.data_source, **self.xr_read_pars)
        else:
            data = self.data_source

        if self.states_sel is not None:
            data = data.isel({self.state_coord: self.states_sel})
        if self.states_loc is not None:
            data = data.sel({self.state_coord: self.states_loc})

        self._N = data.sizes[self.state_coord]
        self._inds = data.coords[self.state_coord].to_numpy()
        if self.format_times_func is not None:
            self._inds = self.format_times_func(self._inds)

        w_name = self.var2col.get(FV.WEIGHT, FV.WEIGHT)
        self._weights = np.zeros((self._N, algo.n_turbines), dtype=FC.DTYPE)
        if w_name in data.data_vars:
            if data[w_name].dims != (self.state_coord,):
                raise ValueError(
                    f"Weights data '{w_name}': Expecting dims ({self.state_coord},), got {data[w_name]}"
                )
            self._weights[:] = data.data_vars[w_name].to_numpy()[:, None]
        elif FV.WEIGHT in self.var2col:
            raise KeyError(
                f"Weight variable '{w_name}' defined in var2col, but not found in data_vars {list(data.data_vars.keys())}"
            )
        else:
            self._weights = np.zeros((self._N, algo.n_turbines), dtype=FC.DTYPE)
            self._weights[:] = 1.0 / self._N

        cols = {}
        self._solo = {}
        for v in self.ovars:
            c = self.var2col.get(v, v)
            if c in self.fixed_vars:
                pass
            elif c in data.attrs:
                self._solo[v] = np.full(self._N, data.attrs)
            elif c in data.data_vars:
                if data[c].dims == (self.state_coord,):
                    self._solo[v] = data.data_vars[c].to_numpy()
                elif data[c].dims == (self.state_coord, self.h_coord):
                    cols[v] = c
                else:
                    raise ValueError(
                        f"Variable '{c}': Expecting dims {(self.state_coord, self.h_coord)}, got {data[c].dims}"
                    )
            else:
                raise KeyError(
                    f"Missing variable '{c}', found data_vars {sorted(list(data.data_vars.keys()))} and attrs {sorted(list(data.attrs.keys()))}"
                )

        if self.heights is not None:
            data = data.sel({self.h_coord: self.heights})
        else:
            self.heights = data[self.h_coord].to_numpy()

        self.H = self.var(FV.H)
        self.VARS = self.var("vars")
        self.DATA = self.var("data")

        idata = States.load_data(self, algo, verbosity)
        idata["coords"][self.H] = self.heights
        idata["coords"][self.VARS] = list(cols.keys())

        dims = (FC.STATE, self.VARS, self.H)
        idata["data_vars"][self.DATA] = (
            dims,
            np.stack(
                [data.data_vars[c].to_numpy() for c in cols.values()], axis=1
            ).astype(FC.DTYPE),
        )

        for v, d in self._solo.items():
            idata["data_vars"][self.var(v)] = ((FC.STATE,), d.astype(FC.DTYPE))
        self._solo = list(self._solo.keys())

        return idata


class MultiHeightTimeseries(MultiHeightStates):
    """
    Multi-height timeseries states data.

    :group: input.states

    """

    RDICT = {"index_col": 0, "parse_dates": [0]}


class MultiHeightNCTimeseries(MultiHeightNCStates):
    """
    Multi-height timeseries from xarray Dataset.

    :group: input.states

    """

    def __init__(
        self,
        *args,
        time_coord=FC.TIME,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        args: tuple, optional
            Parameters for the base class
        time_coord: str
            Name of the state coordinate
        kwargs: dict, optional
            Parameters for the base class

        """
        super().__init__(*args, state_coord=time_coord, **kwargs)



import numpy as np

import foxes.constants as FC
from .rotor_wd import RotorWD
from foxes.core import WakeFrame


class FarmOrder(WakeFrame):
    """
    Invokes turbine ordering as defined
    by the wind farm.

    Warning: This is for testing purposes only, and in general
    only gives correct calculation results when used
    in an iterative algorithm.

    Attributes
    ----------
    base_frame: foxes.core.WakeFrame
        The wake frame from which to start

    :group: models.wake_frames

    """

    def __init__(self, base_frame=RotorWD()):
        """
        Constructor.

        Parameters
        ----------
        base_frame: foxes.core.WakeFrame
            The wake frame from which to start

        """
        super().__init__()
        self.base_frame = base_frame

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            Names of all sub models

        """
        return [self.base_frame]

    def calc_order(self, algo, mdata, fdata):
        """ "
        Calculates the order of turbine evaluation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        order: numpy.ndarray
            The turbine order, shape: (n_states, n_turbines)

        """
        order = np.zeros((fdata.n_states, fdata.n_turbines), dtype=FC.ITYPE)
        order[:] = np.arange(fdata.n_turbines)[None, :]

        return order

    def get_wake_coos(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
    ):
        """
        Calculate wake coordinates of rotor points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order

        Returns
        -------
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)

        """
        return self.base_frame.get_wake_coos(algo, mdata, fdata, tdata, downwind_index)

    def get_centreline_points(self, algo, mdata, fdata, downwind_index, x):
        """
        Gets the points along the centreline for given
        values of x.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The wake frame x coordinates, shape: (n_states, n_points)

        Returns
        -------
        points: numpy.ndarray
            The centreline points, shape: (n_states, n_points, 3)

        """
        return self.base_frame.get_centreline_points(
            algo, mdata, fdata, downwind_index, x
        )



from pathlib import Path

from foxes.data import FARM, StaticData
from .from_json import add_from_json
from .from_csv import add_from_csv


def add_from_file(farm, file_path, *args, verbosity=1, dbook=None, **kwargs):
    """
    Add turbines from file.

    The method is inferred according to the file suffix.

    Parameters
    ----------
    farm: foxes.WindFarm
        The wind farm
    file_path: str
        Path to the file
    verbosity
    args: tuple, optional
        Parameters forwarded to the method
    verbosity: int
        The verbosity level, 0 = silent
    dbook: foxes.DataBook, optional
        The data book, or None for default
    kwargs: dict, optional
        Parameters forwarded to the method

    :group: input.farm_layout

    """

    fpath = Path(file_path)
    dbook = StaticData() if dbook is None else dbook

    if not fpath.is_file():
        if verbosity:
            print(f"Reading static data '{fpath.name}' from context '{FARM}'")
        fpath = dbook.get_file_path(FARM, fpath.name, check_raw=False)

    if fpath.suffix == ".json":
        add_from_json(farm, fpath, *args, **kwargs)
    elif (
        fpath.suffix == ".csv"
        or (len(fpath) > 7 and fpath[-7:] == ".csv.gz")
        or (len(fpath) > 8 and fpath[-8:] == ".csv.bz2")
        or (len(fpath) > 8 and fpath[-8:] == ".csv.zip")
    ):
        add_from_csv(farm, fpath, *args, verbosity=verbosity, **kwargs)
    else:
        raise KeyError(
            f"Unsupported file suffix: '{fpath}'. Please provide any of: json, csv, csv.gz, csv.bz2, csv.zip"
        )



import numpy as np

from foxes.core import FarmDataModel
import foxes.variables as FV


class ReorderFarmOutput(FarmDataModel):
    """
    Reorders final results to state-turbine dimensions

    Attributes
    ----------
    outputs: list of str, optional
        The output variables, or None for defaults

    :group: algorithms.downwind.models

    """

    def __init__(self, outputs):
        """
        Constructor

        Parameters
        ----------
        outputs: list of str, optional
            The output variables, or None for defaults

        """
        super().__init__(pre_rotor=False)
        self.outputs = outputs

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.outputs if self.outputs is not None else algo.farm_vars

    def calculate(self, algo, mdata, fdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        ssel = fdata[FV.ORDER_SSEL]
        order_inv = fdata[FV.ORDER_INV]

        out = {}
        for v in self.output_farm_vars(algo):
            if v != FV.ORDER and np.any(fdata[v] != fdata[v][0, 0, None, None]):
                out[v] = fdata[v][ssel, order_inv]
            else:
                out[v] = fdata[v]

        return out



"""
Induction wake models
"""

from .rankine_half_body import RankineHalfBody
from .rathmann import Rathmann
from .self_similar import SelfSimilar
from .self_similar2020 import SelfSimilar2020
from .vortex_sheet import VortexSheet



import numpy as np
from abc import ABC
from itertools import count
from copy import deepcopy

import foxes.constants as FC
from .data import Data


class Model(ABC):
    """
    Base class for all models.

    Attributes
    ----------
    name: str
        The model name

    :group: core

    """

    _ids = {}

    def __init__(self):
        """
        Constructor.
        """
        t = type(self).__name__
        if t not in self._ids:
            self._ids[t] = count(0)
        self._id = next(self._ids[t])

        self.name = f"{type(self).__name__}"
        if self._id > 0:
            self.name += f"_instance{self._id}"

        self._store = {}
        self.__initialized = False

    def __repr__(self):
        return f"{type(self).__name__}()"

    @property
    def model_id(self):
        """
        Unique id based on the model type.

        Returns
        -------
        int
            Unique id of the model object

        """
        return self._id

    def var(self, v):
        """
        Creates a model specific variable name.

        Parameters
        ----------
        v: str
            The variable name

        Returns
        -------
        str
            Model specific variable name

        """
        return f"{self.name}_{v}"

    @property
    def initialized(self):
        """
        Initialization flag.

        Returns
        -------
        bool :
            True if the model has been initialized.

        """
        return self.__initialized

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            All sub models

        """
        return []

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        return {"coords": {}, "data_vars": {}}

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        if not self.initialized:
            pr = False
            for m in self.sub_models():
                if not m.initialized:
                    if verbosity > 1 and not pr:
                        print(f">> {self.name}: Starting sub-model initialization >> ")
                        pr = True
                    m.initialize(algo, verbosity)
            if pr:
                print(f"<< {self.name}: Finished sub-model initialization << ")

            if verbosity > 0:
                print(f"Initializing model '{self.name}'")
            algo.store_model_data(self, self.load_data(algo, verbosity), force)

            self.__initialized = True

    def finalize(self, algo, verbosity=0):
        """
        Finalizes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        if self.initialized:
            pr = False
            for m in self.sub_models():
                if verbosity > 1 and not pr:
                    print(f">> {self.name}: Starting sub-model finalization >> ")
                    pr = True
                m.finalize(algo, verbosity)
            if pr:
                print(f"<< {self.name}: Finished sub-model finalization << ")

            if verbosity > 0:
                print(f"Finalizing model '{self.name}'")
            algo.del_model_data(self)

            self._store = {}
            self.__initialized = False

    def get_data(
        self,
        variable,
        target,
        lookup="smfp",
        mdata=None,
        fdata=None,
        tdata=None,
        downwind_index=None,
        accept_none=False,
        accept_nan=True,
        algo=None,
        upcast=False,
    ):
        """
        Getter for a data entry in the model object
        or provided data sources

        Parameters
        ----------
        variable: str
            The variable, serves as data key
        target: str, optional
            The dimensions identifier for the output,
            FC.STATE_TURBINE, FC.STATE_TARGET or
            FC.STATE_TARGET_TPOINT
        lookup: str
            The order of data sources. Combination of:
            's' for self,
            'm' for mdata,
            'f' for fdata,
            't' for tdata,
            'w' for wake modelling data
        mdata: foxes.core.Data, optional
            The model data
        fdata: foxes.core.Data, optional
            The farm data
        tdata: foxes.core.Data, optional
            The target point data
        downwind_index: int, optional
            The index in the downwind order
        data_prio: bool
            First search the data source, then the object
        accept_none: bool
            Do not throw an error if data entry is None
        accept_nan: bool
            Do not throw an error if data entry is np.nan
        algo: foxes.core.Algorithm, optional
            The algorithm, needed for data from previous iteration
        upcast: bool
            Flag for ensuring targets dimension,
            otherwise dimension 1 is entered

        """

        def _geta(a):
            sources = [s for s in [mdata, fdata, tdata, algo, self] if s is not None]
            for s in sources:
                try:
                    if a == "states_i0":
                        out = s.states_i0(counter=True, algo=algo)
                        if out is not None:
                            return out
                    else:
                        out = getattr(s, a)
                        if out is not None:
                            return out
                except AttributeError:
                    pass
            raise KeyError(
                f"Model '{self.name}': Failed to determine '{a}'. Maybe add to arguments of get_data: mdata, fdata, tdata, algo?"
            )

        n_states = _geta("n_states")
        if target == FC.STATE_TURBINE:
            n_turbines = _geta("n_turbines")
            dims = (FC.STATE, FC.TURBINE)
            shp = (n_states, n_turbines)
        elif target == FC.STATE_TARGET:
            n_targets = _geta("n_targets")
            dims = (FC.STATE, FC.TARGET)
            shp = (n_states, n_targets)
        elif target == FC.STATE_TARGET_TPOINT:
            n_targets = _geta("n_targets")
            n_tpoints = _geta("n_tpoints")
            dims = (FC.STATE, FC.TARGET, FC.TPOINT)
            shp = (n_states, n_targets, n_tpoints)
        else:
            raise KeyError(
                f"Model '{self.name}': Wrong parameter 'target = {target}'. Choices: {FC.STATE_TURBINE}, {FC.STATE_TARGET}, {FC.STATE_TARGET_TPOINT}"
            )

        out = None
        out_dims = None
        for s in lookup:
            # lookup self:
            if s == "s" and hasattr(self, variable):
                a = getattr(self, variable)
                if a is not None:
                    if not upcast:
                        out = a
                        out_dims = None
                    elif target == FC.STATE_TURBINE:
                        out = np.full((n_states, n_turbines), np.nan, dtype=FC.DTYPE)
                        out[:] = a
                        out_dims = (FC.STATE, FC.TURBINE)
                    elif target == FC.STATE_TARGET:
                        out = np.full((n_states, n_targets), np.nan, dtype=FC.DTYPE)
                        out[:] = a
                        out_dims = (FC.STATE, FC.TARGET)
                    elif target == FC.STATE_TARGET_TPOINT:
                        out = np.full(
                            (n_states, n_targets, n_tpoints), np.nan, dtype=FC.DTYPE
                        )
                        out[:] = a
                        out_dims = (FC.STATE, FC.TARGET, FC.TPOINT)
                    else:
                        raise NotImplementedError

            # lookup mdata:
            elif (
                s == "m"
                and mdata is not None
                and variable in mdata
                and tuple(mdata.dims[variable]) == dims
            ):
                out = mdata[variable]
                out_dims = dims

            # lookup fdata:
            elif (
                s == "f"
                and fdata is not None
                and variable in fdata
                and tuple(fdata.dims[variable]) == (FC.STATE, FC.TURBINE)
            ):
                out = fdata[variable]
                out_dims = (FC.STATE, FC.TURBINE)

            # lookup pdata:
            elif (
                s == "t"
                and tdata is not None
                and variable in tdata
                and tuple(tdata.dims[variable]) == (FC.STATE, FC.TARGET, FC.TPOINT)
            ):
                out = tdata[variable]
                out_dims = (FC.STATE, FC.TARGET, FC.TPOINT)

            # lookup wake modelling data:
            elif (
                s == "w"
                and fdata is not None
                and tdata is not None
                and variable in fdata
                and tuple(fdata.dims[variable]) == (FC.STATE, FC.TURBINE)
                and downwind_index is not None
                and algo is not None
            ):
                out, out_dims = algo.wake_frame.get_wake_modelling_data(
                    algo,
                    variable,
                    downwind_index,
                    fdata,
                    tdata=tdata,
                    target=target,
                    upcast=upcast,
                )

            if out is not None:
                break

        # cast dimensions:
        if out_dims != dims:
            if out_dims is None:
                if upcast:
                    out0 = out
                    out = np.zeros(shp, dtype=FC.DTYPE)
                    out[:] = out0
                    out_dims = dims
                    del out0
                else:
                    out_dims = tuple([1 for _ in dims])

            elif out_dims == (FC.STATE, FC.TURBINE):
                if downwind_index is None:
                    raise KeyError(
                        f"Require downwind_index for target {target} and out dims {out_dims}"
                    )
                out0 = out[:, downwind_index, None]
                if len(dims) == 3:
                    out0 = out0[:, :, None]
                if upcast:
                    out = np.zeros(shp, dtype=FC.DTYPE)
                    out[:] = out0
                    out_dims = dims
                else:
                    out = out0
                    out_dims = (FC.STATE, 1) if len(dims) == 2 else (FC.STATE, 1, 1)
                del out0

            elif out_dims == (FC.STATE, 1):
                out0 = out
                if len(dims) == 3:
                    out0 = out0[:, :, None]
                    out_dims = (FC.STATE, 1, 1)
                if upcast:
                    out = np.zeros(shp, dtype=FC.DTYPE)
                    out[:] = out0
                    out_dims = dims
                else:
                    out = out0
                del out0

            elif out_dims == (FC.STATE, 1, 1):
                out0 = out
                if len(dims) == 2:
                    out0 = out0[:, :, 0]
                    out_dims = (FC.STATE, 1)
                if upcast:
                    out = np.zeros(shp, dtype=FC.DTYPE)
                    out[:] = out0
                    out_dims = dims
                else:
                    out = out0
                del out0

            else:
                raise NotImplementedError(
                    f"No casting implemented for target {target} and out dims {out_dims} fo upcast {upcast}"
                )

        # data from other chunks, only with iterations:
        if (
            target in [FC.STATE_TARGET, FC.STATE_TARGET_TPOINT]
            and fdata is not None
            and variable in fdata
            and tdata is not None
            and FC.STATES_SEL in tdata
        ):
            if out_dims != dims:
                raise ValueError(
                    f"Model '{self.name}': Iteration data found for variable '{variable}', but missing upcast: out_dims = {out_dims}, expecting {dims}"
                )
            if downwind_index is None:
                raise KeyError(
                    f"Model '{self.name}': Require downwind_index for obtaining results from previous iteration"
                )
            if tdata[FC.STATE_SOURCE_ORDERI] != downwind_index:
                raise ValueError(
                    f"Model '{self.name}': Expecting downwind_index {tdata[FC.STATE_SOURCE_ORDERI]}, got {downwind_index}"
                )
            if algo is None:
                raise ValueError(
                    f"Model '{self.name}': Iteration data found for variable '{variable}', requiring algo"
                )

            i0 = _geta("states_i0")
            sts = tdata[FC.STATES_SEL]
            if target == FC.STATE_TARGET and tdata.n_tpoints != 1:
                # find the mean index and round it to nearest integer:
                sts = tdata.tpoint_mean(FC.STATES_SEL)[:, :, None]
                sts = (sts + 0.5).astype(FC.ITYPE)
            sel = sts < i0
            if np.any(sel):
                if not hasattr(algo, "prev_farm_results"):
                    raise KeyError(
                        f"Model '{self.name}': Iteration data found for variable '{variable}', requiring iterative algorithm"
                    )
                prev_fres = getattr(algo, "prev_farm_results")
                if prev_fres is not None:
                    prev_data = prev_fres[variable].to_numpy()[sts[sel], downwind_index]
                    if target == FC.STATE_TARGET:
                        out[sel[:, :, 0]] = prev_data
                    else:
                        out[sel] = prev_data
                    del prev_fres, prev_data

        # check for None:
        if not accept_none and out is None:
            raise ValueError(
                f"Model '{self.name}': Variable '{variable}' is requested but not found."
            )

        # check for nan:
        elif not accept_nan:
            try:
                if np.all(np.isnan(np.atleast_1d(out))):
                    raise ValueError(
                        f"Model '{self.name}': Requested variable '{variable}' contains NaN values."
                    )
            except TypeError:
                pass

        return out

    def data_to_store(self, name, algo, data):
        """
        Adds data from mdata to the local store, intended
        for iterative runs.

        Parameters
        ----------
        name: str
            The data name
        algo: foxes.core.Algorithm
            The algorithm
        data: foxes.utils.Data
            The mdata, fdata or pdata object

        """
        i0 = data.states_i0(counter=True, algo=algo)
        if i0 not in self._store:
            self._store[i0] = Data(
                data={}, dims={}, loop_dims=data.loop_dims, name=f"{self.name}_{i0}"
            )

        self._store[i0][name] = deepcopy(data[name])
        self._store[i0].dims[name] = (
            deepcopy(data.dims[name]) if name in data.dims else None
        )

    def from_data_or_store(self, name, algo, data, ret_dims=False, safe=False):
        """
        Get data from mdata or local store

        Parameters
        ----------
        name: str
            The data name
        algo: foxes.core.Algorithm
            The algorithm
        data: foxes.utils.Data
            The mdata, fdata or pdata object
        ret_dims: bool
            Return dimensions
        safe: bool
            Return None instead of error if
            not found

        Returns
        -------
        data: numpy.ndarray
            The data
        dims: tuple of dims, optional
            The data dimensions

        """
        if name in data:
            return (data[name], data.dims[name]) if ret_dims else data[name]

        i0 = data.states_i0(counter=True, algo=algo)
        if not safe or (i0 in self._store and name in self._store[i0]):
            if ret_dims:
                return self._store[i0][name], self._store[i0].dims[name]
            else:
                return self._store[i0][name]
        else:
            return (None, None) if ret_dims else None



import foxes.constants as FC
from foxes.utils import all_subclasses

from .turbine_model import TurbineModel


class TurbineType(TurbineModel):
    """
    Abstract base class for turbine type models.

    Rotor diameter and hub height can be overwritten
    by individual settings in the Turbine object.

    Attributes
    ----------
    name: str
        The model name
    D: float
        The rotor diameter
    H: float
        The hub height
    P_nominal: float
        The nominal power in kW
    P_unit: str
        The unit of power

    :group: core

    """

    def __init__(self, name=None, D=None, H=None, P_nominal=None, P_unit="kW"):
        """
        Constructor.

        Parameters
        ----------
        name: str, optional
            The model name
        D: float, optional
            The rotor diameter
        H: float, optional
            The hub height
        P_nominal: float, optional
            The nominal power in kW
        P_unit: str
            The unit of power, choices:
            W, kW, MW, GW

        """
        super().__init__()

        self.name = name if name is not None else type(self).__name__
        self.D = D
        self.H = H
        self.P_nominal = P_nominal
        self.P_unit = P_unit

        if P_unit not in FC.P_UNITS:
            raise KeyError(
                f"Turbine type '{self.name}': Unkown P_unit '{P_unit}', expecting {list(FC.P_UNITS.keys())}"
            )

    def __repr__(self):
        a = f"D={self.D}, H={self.H}, P_nominal={self.P_nominal}, P_unit={self.P_unit}"
        return f"{type(self).__name__}({a})"

    def modify_cutin(self, modify_ct, modify_P):
        """
        Modify the data such that a discontinuity
        at cutin wind speed is avoided

        Parameters
        ----------
        variable: str
            The target variable
        modify_ct: bool
            Flag for modification of the ct curve
        modify_P: bool
            Flag for modification of the power curve

        """
        if modify_ct or modify_P:
            raise NotImplementedError(
                f"Turbine type '{self.name}': Continuous cutin not implemented for modify_ct = {modify_ct}, modify_P = {modify_P}"
            )

    @classmethod
    def new(cls, ttype_type, *args, **kwargs):
        """
        Run-time turbine type factory.

        Parameters
        ----------
        ttype_type: str
            The selected derived class name
        args: tuple, optional
            Additional parameters for constructor
        kwargs: dict, optional
            Additional parameters for constructor

        """

        if ttype_type is None:
            return None

        allc = all_subclasses(cls)
        found = ttype_type in [scls.__name__ for scls in allc]

        if found:
            for scls in allc:
                if scls.__name__ == ttype_type:
                    return scls(*args, **kwargs)

        else:
            estr = "Turbine type class '{}' is not defined, available types are \n {}".format(
                ttype_type, sorted([i.__name__ for i in allc])
            )
            raise KeyError(estr)



import numpy as np

from foxes.opt.core.farm_objective import FarmObjective
import foxes.constants as FC


class MaxNTurbines(FarmObjective):
    """
    Maximizes the number of turrbines.

    Attributes
    ----------
    check_valid: bool
        Check FC.VALID variable before counting

    :group: opt.objectives

    """

    def __init__(
        self,
        problem,
        name="max_n_turbines",
        check_valid=True,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        problem: foxes.opt.FarmOptProblem
            The underlying optimization problem
        name: str
            The name of the objective function
        check_valid: bool
            Check FC.VALID variable before counting
        kwargs: dict, optional
            Additional parameters for `FarmObjective`

        """
        super().__init__(problem, name, **kwargs)
        self.check_valid = check_valid

    def n_components(self):
        """
        Returns the number of components of the
        function.

        Returns
        -------
        int:
            The number of components.

        """
        return 1

    def maximize(self):
        """
        Returns flag for maximization of each component.

        Returns
        -------
        flags: np.array
            Bool array for component maximization,
            shape: (n_components,)

        """
        return [True]

    def calc_individual(self, vars_int, vars_float, problem_results, components=None):
        """
        Calculate values for a single individual of the
        underlying problem.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float: np.array
            The float variable values, shape: (n_vars_float,)
        problem_results: Any
            The results of the variable application
            to the problem
        components: list of int, optional
            The selected components or None for all

        Returns
        -------
        values: np.array
            The component values, shape: (n_sel_components,)

        """
        if FC.VALID in problem_results and self.check_valid:
            vld = np.sum(problem_results[FC.VALID].to_numpy(), axis=1)
            if np.min(vld) != np.max(vld):
                raise ValueError(
                    f"Objective '{self.name}': Number of valid turbines is state dependend, counting impossible"
                )
            return np.array([vld[0]], dtype=np.float64)
        else:
            return np.array([self.farm.n_turbines], dtype=np.float64)

    def calc_population(self, vars_int, vars_float, problem_results, components=None):
        """
        Calculate values for all individuals of a population.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float: np.array
            The float variable values, shape: (n_pop, n_vars_float)
        problem_results: Any
            The results of the variable application
            to the problem
        components: list of int, optional
            The selected components or None for all

        Returns
        -------
        values: np.array
            The component values, shape: (n_pop, n_sel_components)

        """
        n_pop = problem_results["n_pop"].to_numpy()
        if self.check_valid:
            n_states = problem_results["n_org_states"].to_numpy()
            n_turbines = self.farm.n_turbines
            vld = (
                problem_results[FC.VALID]
                .to_numpy()
                .reshape(n_pop, n_states, n_turbines)
            )
            vld = np.sum(vld, axis=2)
            if np.any(np.min(vld, axis=1) != np.max(vld, axis=1)):
                raise ValueError(
                    f"Objective '{self.name}': Number of valid turbines is state dependend, counting impossible"
                )
            return vld[:, 0, None]
        else:
            return np.full((n_pop, 1), self.farm.n_turbines, dtype=vars_float.dtype)



import numpy as np

from foxes.core import WakeSuperposition
import foxes.variables as FV
import foxes.constants as FC


class WSQuadratic(WakeSuperposition):
    """
    Quadratic supersposition of wind deficit results

    Attributes
    ----------
    scale_amb: bool
        Flag for scaling wind deficit with ambient wind speed
        instead of waked wind speed
    lim_low: float
        Lower limit of the final waked wind speed
    lim_high: float
        Upper limit of the final waked wind speed

    :group: models.wake_superpositions

    """

    def __init__(self, scale_amb=False, lim_low=None, lim_high=None):
        """
        Constructor.

        Parameters
        ----------
        scale_amb: bool
            Flag for scaling wind deficit with ambient wind speed
            instead of waked wind speed
        lim_low: float
            Lower limit of the final waked wind speed
        lim_high: float
            Upper limit of the final waked wind speed

        """
        super().__init__()

        self.scale_amb = scale_amb
        self.lim_low = lim_low
        self.lim_high = lim_high

    def __repr__(self):
        a = f"scale_amb={self.scale_amb}, lim_low={self.lim_low}, lim_high={self.lim_high}"
        return f"{type(self).__name__}({a})"

    def input_farm_vars(self, algo):
        """
        The variables which are needed for running
        the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        input_vars: list of str
            The input variable names

        """
        return [FV.AMB_REWS] if self.scale_amb else [FV.REWS]

    def add_wake(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        st_sel,
        variable,
        wake_delta,
        wake_model_result,
    ):
        """
        Add a wake delta to previous wake deltas,
        at rotor points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order
        st_sel: numpy.ndarray of bool
            The selection of targets, shape: (n_states, n_targets)
        variable: str
            The variable name for which the wake deltas applies
        wake_delta: numpy.ndarray
            The original wake deltas, shape:
            (n_states, n_targets, n_tpoints, ...)
        wake_model_result: numpy.ndarray
            The new wake deltas of the selected rotors,
            shape: (n_st_sel, n_tpoints, ...)

        Returns
        -------
        wdelta: numpy.ndarray
            The updated wake deltas, shape:
            (n_states, n_targets, n_tpoints, ...)

        """
        if variable not in [FV.REWS, FV.REWS2, FV.REWS3, FV.WS]:
            raise ValueError(
                f"Superposition '{self.name}': Expecting wind speed variable, got {variable}"
            )

        if np.any(st_sel):
            scale = self.get_data(
                FV.AMB_REWS if self.scale_amb else FV.REWS,
                FC.STATE_TARGET,
                lookup="w",
                algo=algo,
                fdata=fdata,
                tdata=tdata,
                downwind_index=downwind_index,
                upcast=True,
            )[st_sel, None]

            wake_delta[st_sel] += (scale * wake_model_result) ** 2

        return wake_delta

    def calc_final_wake_delta(
        self,
        algo,
        mdata,
        fdata,
        variable,
        amb_results,
        wake_delta,
    ):
        """
        Calculate the final wake delta after adding all
        contributions.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        variable: str
            The variable name for which the wake deltas applies
        amb_results: numpy.ndarray
            The ambient results at targets,
            shape: (n_states, n_targets, n_tpoints)
        wake_delta: numpy.ndarray
            The wake deltas at targets, shape:
            (n_states, n_targets, n_tpoints)

        Returns
        -------
        final_wake_delta: numpy.ndarray
            The final wake delta, which will be added to the ambient
            results by simple plus operation. Shape:
            (n_states, n_targets, n_tpoints)

        """
        w = -np.sqrt(wake_delta)
        if self.lim_low is not None:
            w = np.maximum(w, self.lim_low - amb_results)
        if self.lim_high is not None:
            w = np.minimum(w, self.lim_high - amb_results)
        return w


class WSQuadraticLocal(WakeSuperposition):
    """
    Local quadratic supersposition of wind deficit results

    Attributes
    ----------
    lim_low: float
        Lower limit of the final waked wind speed
    lim_high: float
        Upper limit of the final waked wind speed

    :group: models.wake_superpositions

    """

    def __init__(self, lim_low=None, lim_high=None):
        """
        Constructor.

        Parameters
        ----------
        lim_low: float
            Lower limit of the final waked wind speed
        lim_high: float
            Upper limit of the final waked wind speed

        """
        super().__init__()
        self.lim_low = lim_low
        self.lim_high = lim_high

    def __repr__(self):
        a = f"lim_low={self.lim_low}, lim_high={self.lim_high}"
        return f"{type(self).__name__}({a})"

    def input_farm_vars(self, algo):
        """
        The variables which are needed for running
        the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        input_vars: list of str
            The input variable names

        """
        return []

    def add_wake(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        st_sel,
        variable,
        wake_delta,
        wake_model_result,
    ):
        """
        Add a wake delta to previous wake deltas,
        at rotor points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order
        st_sel: numpy.ndarray of bool
            The selection of targets, shape: (n_states, n_targets)
        variable: str
            The variable name for which the wake deltas applies
        wake_delta: numpy.ndarray
            The original wake deltas, shape:
            (n_states, n_targets, n_tpoints, ...)
        wake_model_result: numpy.ndarray
            The new wake deltas of the selected rotors,
            shape: (n_st_sel, n_tpoints, ...)

        Returns
        -------
        wdelta: numpy.ndarray
            The updated wake deltas, shape:
            (n_states, n_targets, n_tpoints, ...)

        """
        if variable not in [FV.REWS, FV.REWS2, FV.REWS3, FV.WS]:
            raise ValueError(
                f"Superposition '{self.name}': Expecting wind speed variable, got {variable}"
            )

        if np.any(st_sel):
            wake_delta[st_sel] += wake_model_result**2

        return wake_delta

    def calc_final_wake_delta(
        self,
        algo,
        mdata,
        fdata,
        variable,
        amb_results,
        wake_delta,
    ):
        """
        Calculate the final wake delta after adding all
        contributions.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        variable: str
            The variable name for which the wake deltas applies
        amb_results: numpy.ndarray
            The ambient results at targets,
            shape: (n_states, n_targets, n_tpoints)
        wake_delta: numpy.ndarray
            The wake deltas at targets, shape:
            (n_states, n_targets, n_tpoints)

        Returns
        -------
        final_wake_delta: numpy.ndarray
            The final wake delta, which will be added to the ambient
            results by simple plus operation. Shape:
            (n_states, n_targets, n_tpoints)

        """
        w = -np.sqrt(wake_delta) * amb_results
        if self.lim_low is not None:
            w = np.maximum(w, self.lim_low - amb_results)
        if self.lim_high is not None:
            w = np.minimum(w, self.lim_high - amb_results)
        return w



from foxes.core import PointDataModel
import foxes.variables as FV


class WakeDeltas(PointDataModel):
    """
    This point model simply subtracts ambient results
    from waked results.

    Attributes
    ----------
    vars: list of str
        The variables
    normalize: bool
        Divide resulting deltas by ambient values

    :group: models.point_models

    """

    def __init__(self, vars, normalize=False):
        """
        Constructor.

        Parameters
        ----------
        vars: list of str
            The variables
        normalize: bool
            Divide resulting deltas by ambient values

        """
        super().__init__()
        self.vars = vars
        self.normalize = normalize

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return [f"DELTA_{v}" for v in self.vars]

    def calculate(self, algo, mdata, fdata, pdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_points)

        """

        out = {f"DELTA_{v}": pdata[v] - pdata[FV.var2amb[v]] for v in self.vars}

        if self.normalize:
            for v in self.vars:
                out[v] /= pdata[FV.var2amb[v]]

        return out



import json
import numpy as np
from copy import deepcopy

from foxes.core import Turbine


def add_from_json(
    farm, file_path, set_farm_name=True, verbosity=1, **turbine_parameters
):
    """
    Add turbimes from a json file.

    Parameters
    ----------
    farm: foxes.WindFarm
        The wind farm
    file_path: str
        Path to the file
    set_farm_name: bool
        Flag for inferring wind farm name from data
    verbosity: int
        The verbosity level, 0 = silent
    turbine_parameters: dict, optional
        Parameters forwarded to `foxes.core.Turbine`

    :group: input.farm_layout

    """

    if verbosity:
        print("Reading file", file_path)
    with open(file_path) as f:
        dict = json.load(f)

    keys = list(dict.keys())
    if len(keys) != 1:
        raise KeyError("Only one wind farm supported by flappy at the moment.")

    farm_name = keys[0]
    fdict = dict[farm_name]

    if set_farm_name:
        farm.name = farm_name

    for wt_name, wdict in fdict.items():
        pars = deepcopy(turbine_parameters)
        if "D" in wdict:
            pars["D"] = wdict["D"]
        if "H" in wdict:
            pars["H"] = wdict["H"]
        if "turbine_models" in wdict:
            pars["turbine_models"] = wdict["turbine_models"] + pars.get(
                "turbine_models", []
            )

        wdict = fdict[wt_name]
        farm.add_turbine(
            Turbine(
                xy=np.array([wdict["UTMX"], wdict["UTMY"]]),
                index=wdict.get("id", None),
                name=wt_name,
                **pars
            ),
            verbosity=verbosity,
        )



from .set_amb_farm_results import SetAmbFarmResults
from .set_amb_point_results import SetAmbPointResults
from .farm_wakes_calc import FarmWakesCalculation
from .point_wakes_calc import PointWakesCalculation
from .init_farm_data import InitFarmData
from .reorder_farm_output import ReorderFarmOutput



import numpy as np
from copy import deepcopy

from foxes.opt.core import FarmVarsProblem, FarmOptProblem
from foxes.models.turbine_models import Calculator
import foxes.variables as FV
import foxes.constants as FC


class RegularLayoutOptProblem(FarmVarsProblem):
    """
    Places turbines on a regular grid and optimizes
    its parameters.

    Attributes
    ----------
    min_spacing: float
        The minimal turbine spacing

    :group: opt.problems.layout

    """

    SPACING_X = "spacing_x"
    SPACING_Y = "spacing_y"
    OFFSET_X = "offset_X"
    OFFSET_Y = "offset_Y"
    ANGLE = "angle"

    def __init__(
        self,
        name,
        algo,
        min_spacing,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        name: str
            The problem's name
        algo: foxes.core.Algorithm
            The algorithm
        min_spacing: float
            The minimal turbine spacing
        kwargs: dict, optional
            Additional parameters for `FarmVarsProblem`

        """
        super().__init__(name, algo, **kwargs)
        self.min_spacing = min_spacing

    def initialize(self, verbosity=1, **kwargs):
        """
        Initialize the object.

        Parameters
        ----------
        verbosity: int
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Additional parameters for super class init

        """
        self._mname = self.name + "_calc"
        for t in self.algo.farm.turbines:
            if self._mname not in t.models:
                t.models.append(self._mname)
        self._turbine = deepcopy(self.farm.turbines[-1])

        self.algo.mbook.turbine_models[self._mname] = Calculator(
            in_vars=[FC.VALID, FV.P, FV.CT],
            out_vars=[FC.VALID, FV.P, FV.CT],
            func=lambda valid, P, ct, st_sel: (valid, P * valid, ct * valid),
            pre_rotor=False,
        )

        b = self.farm.boundary
        assert b is not None, f"Problem '{self.name}': Missing wind farm boundary."
        pmax = b.p_max()
        pmin = b.p_min()
        self._pmin = pmin
        self._xy0 = 0.5 * (pmin + pmax)
        self._halfspan = (pmax - pmin) / 2
        self._halflen = np.linalg.norm(self._halfspan)
        self.max_spacing = 2 * (self._halflen + self.min_spacing)
        self._halfn = int(self._halflen / self.min_spacing)
        if self._halfn * self.min_spacing < self._halflen:
            self._halfn += 1
        self._nrow = 2 * self._halfn + 1
        self._nturb = self._nrow**2

        if verbosity > 0:
            print(f"Problem '{self.name}':")
            print(f"  xy0          = {self._xy0}")
            print(f"  span         = {np.linalg.norm(self._halfspan*2):.2f}")
            print(f"  min spacing  = {self.min_spacing:.2f}")
            print(f"  max spacing  = {self.max_spacing:.2f}")
            print(f"  n row turbns = {self._nrow}")
            print(f"  n turbines   = {self._nturb}")
            print(f"  turbine mdls = {self._turbine.models}")

        if self.farm.n_turbines < self._nturb:
            for i in range(self._nturb - self.farm.n_turbines):
                ti = len(self.farm.turbines)
                self.farm.turbines.append(deepcopy(self._turbine))
                self.farm.turbines[-1].index = ti
                self.farm.turbines[-1].name = f"T{ti}"
        elif self.farm.n_turbines > self._nturb:
            self.farm.turbines = self.farm.turbines[: self._nturb]
        self.algo.n_turbines = self._nturb

        super().initialize(
            pre_rotor_vars=[FV.X, FV.Y, FC.VALID],
            post_rotor_vars=[],
            verbosity=verbosity,
            **kwargs,
        )

    def var_names_float(self):
        """
        The names of float variables.

        Returns
        -------
        names: list of str
            The names of the float variables

        """
        return [
            self.SPACING_X,
            self.SPACING_Y,
            self.OFFSET_X,
            self.OFFSET_Y,
            self.ANGLE,
        ]

    def initial_values_float(self):
        """
        The initial values of the float variables.

        Returns
        -------
        values: numpy.ndarray
            Initial float values, shape: (n_vars_float,)

        """
        return [self.min_spacing, self.min_spacing, 0.0, 0.0, 0.0]

    def min_values_float(self):
        """
        The minimal values of the float variables.

        Use -numpy.inf for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Minimal float values, shape: (n_vars_float,)

        """
        return [
            self.min_spacing,
            self.min_spacing,
            -self._halfspan[0] - self.min_spacing,
            -self._halfspan[1] - self.min_spacing,
            0.0,
        ]

    def max_values_float(self):
        """
        The maximal values of the float variables.

        Use numpy.inf for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Maximal float values, shape: (n_vars_float,)

        """
        return [
            self.max_spacing,
            self.max_spacing,
            self._halfspan[0] + self.min_spacing,
            self._halfspan[1] + self.min_spacing,
            90.0,
        ]

    def opt2farm_vars_individual(self, vars_int, vars_float):
        """
        Translates optimization variables to farm variables

        Parameters
        ----------
        vars_int: numpy.ndarray
            The integer optimization variable values,
            shape: (n_vars_int,)
        vars_float: numpy.ndarray
            The float optimization variable values,
            shape: (n_vars_float,)

        Returns
        -------
        farm_vars: dict
            The foxes farm variables. Key: var name,
            value: numpy.ndarray with values, shape:
            (n_states, n_sel_turbines)

        """

        dx, dy, ox, oy, a = vars_float
        n_states = self.algo.n_states
        nx = self._nrow
        ny = self._nrow

        a = np.deg2rad(a)
        nax = np.array([np.cos(a), np.sin(a), 0.0], dtype=FC.DTYPE)
        nay = np.cross(np.array([0.0, 0.0, 1.0], dtype=FC.DTYPE), nax)
        x0 = self._xy0 + np.array([ox, oy], dtype=FC.DTYPE)

        pts = np.zeros((n_states, nx, ny, 2), dtype=FC.DTYPE)
        pts[:] = (
            x0[None, None, None, :]
            + np.arange(nx)[None, :, None, None] * dx * nax[None, None, None, :2]
            + np.arange(ny)[None, None, :, None] * dy * nay[None, None, None, :2]
        )

        pts = pts.reshape(n_states, nx * ny, 2)
        valid = self.farm.boundary.points_inside(pts.reshape(n_states * nx * ny, 2))

        farm_vars = {
            FV.X: pts[:, :, 0],
            FV.Y: pts[:, :, 1],
            FC.VALID: valid.reshape(n_states, nx * ny).astype(FC.DTYPE),
        }

        return farm_vars

    def opt2farm_vars_population(self, vars_int, vars_float, n_states):
        """
        Translates optimization variables to farm variables

        Parameters
        ----------
        vars_int: numpy.ndarray
            The integer optimization variable values,
            shape: (n_pop, n_vars_int)
        vars_float: numpy.ndarray
            The float optimization variable values,
            shape: (n_pop, n_vars_float)
        n_states: int
            The number of original (non-pop) states

        Returns
        -------
        farm_vars: dict
            The foxes farm variables. Key: var name,
            value: numpy.ndarray with values, shape:
            (n_pop, n_states, n_sel_turbines)

        """
        n_pop = len(vars_float)
        n_turbines = self.farm.n_turbines
        dx = vars_float[:, 0]
        dy = vars_float[:, 1]
        ox = vars_float[:, 2]
        oy = vars_float[:, 3]
        nx = self._nrow
        ny = self._nrow
        a = vars_float[:, 4]
        N = self._nturb

        a = np.deg2rad(a)
        nax = np.stack([np.cos(a), np.sin(a), np.zeros_like(a)], axis=-1)
        naz = np.zeros_like(nax)
        naz[..., 2] = 1
        nay = np.cross(naz, nax)

        pts = np.zeros((n_pop, n_states, nx, ny, 2), dtype=FC.DTYPE)
        pts[:] = self._xy0[None, None, None, None, :]
        pts[..., 0] += ox[:, None, None, None]
        pts[..., 1] += oy[:, None, None, None]
        pts[:] += (
            np.arange(nx)[None, None, :, None, None]
            * dx[:, None, None, None, None]
            * nax[:, None, None, None, :2]
            + np.arange(ny)[None, None, None, :, None]
            * dy[:, None, None, None, None]
            * nay[:, None, None, None, :2]
        )

        qts = np.zeros((n_pop, n_states, n_turbines, 2))
        qts[:, :N] = pts.reshape(n_pop, n_states, N, 2)
        del pts

        valid = self.farm.boundary.points_inside(
            qts.reshape(n_pop * n_states * n_turbines, 2)
        )

        farm_vars = {
            FV.X: qts[:, :, :, 0],
            FV.Y: qts[:, :, :, 1],
            FC.VALID: valid.reshape(n_pop, n_states, n_turbines).astype(FC.DTYPE),
        }

        return farm_vars

    def finalize_individual(self, vars_int, vars_float, verbosity=1):
        """
        Finalization, given the champion data.

        Parameters
        ----------
        vars_int: np.array
            The optimal integer variable values, shape: (n_vars_int,)
        vars_float: np.array
            The optimal float variable values, shape: (n_vars_float,)
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        problem_results: Any
            The results of the variable application
            to the problem
        objs: np.array
            The objective function values, shape: (n_objectives,)
        cons: np.array
            The constraints values, shape: (n_constraints,)

        """
        farm_vars = self.opt2farm_vars_individual(vars_int, vars_float)
        sel = np.where(farm_vars[FC.VALID][0])[0]
        x = farm_vars[FV.X][0, sel]
        y = farm_vars[FV.Y][0, sel]

        self.farm.turbines = [t for i, t in enumerate(self.farm.turbines) if i in sel]
        for i, t in enumerate(self.farm.turbines):
            t.xy = np.array([x[i], y[i]], dtype=FC.DTYPE)
            t.models = [m for m in t.models if m not in [self.name, self._mname]]
            t.index = i
            t.name = f"T{i}"
        self.algo.update_n_turbines()

        return FarmOptProblem.finalize_individual(
            self, vars_int, vars_float, verbosity=1
        )



from tabnanny import verbose
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from foxes.input.states import StatesTable
from foxes.core import WindFarm, Turbine
from foxes.algorithms import Downwind
from foxes.models.turbine_models import SetFarmVars
import foxes.variables as FV
import foxes.constants as FC
from .output import Output


class TurbineTypeCurves(Output):
    """
    Creates power and ct curves for turbine
    types, optionally including derating/boost.

    Attributes
    ----------
    mbook: foxes.models.ModelBook
        The model book

    :group: output

    """

    def __init__(self, mbook):
        """
        Constructor.

        Parameters
        ----------
        mbook: foxes.models.ModelBook
            The model book

        """
        self.mbook = mbook

    def plot_curves(
        self,
        turbine_type,
        variables,
        P_max=None,
        titles=None,
        x_label=None,
        y_labels=None,
        ws_min=0.0,
        ws_max=30.0,
        ws_step=0.1,
        ti=0.05,
        rho=1.225,
        axs=None,
        figsize=None,
        pmax_args={},
        **kwargs,
    ):
        """
        Plot the power or ct curve.

        Parameters
        ----------
        turbine_type: str
            The turbine type name from the
            model book
        variables: str or list of str
            For example FV.P or FV.CT
        P_max: float, optional
            The power mask value, if of interest
        titles: list of str, optional
            The plot titles, one for each variable
        x_label: str, optional
            The x axis label
        y_labels: list of str, optional
            The y axis lables, one for each variable
        ws_min: float
            The minimal wind speed
        ws_max: float
            The maximal wind speed
        ws_step: float
            The wind speed step size
        ti: float
            The TI value
        rho: float
            The air density value
        axs: list of pyplot.Axis, optional
            The axis, one for each variable
        figsize: tuple
            The figsize argument for plt.subplots()
            in case ax is not provided
        pmax_args: dict, optional
            Additionals parameters for plt.plot()
            for power mask case
        kwargs: dict, optional
            Additional parameters for plt.plot()

        Returns
        -------
        axs: list of pyplot.Axis
            The plot axes, one for each variable

        """
        vars = [variables] if isinstance(variables, str) else variables
        if isinstance(titles, str):
            titles = [titles]
        elif titles is None:
            titles = [None for v in vars]
        if isinstance(y_labels, str):
            y_labels = [y_labels]
        elif y_labels is None:
            y_labels = [None for v in vars]
        if not isinstance(axs, (list, tuple, np.ndarray)):
            axs = [axs]

        ws = np.arange(ws_min, ws_max + ws_step, ws_step, dtype=FC.DTYPE)
        n_states = len(ws)
        sdata = pd.DataFrame(index=range(n_states))
        sdata.index.name = FC.STATE
        sdata[FV.WS] = ws

        models = [turbine_type]

        states = StatesTable(
            sdata,
            output_vars={FV.WS, FV.WD, FV.TI, FV.RHO},
            fixed_vars={FV.WD: 270.0, FV.TI: ti, FV.RHO: rho},
        )

        farm = WindFarm()
        farm.add_turbine(
            Turbine(xy=[0.0, 0.0], turbine_models=models),
            verbosity=0,
        )

        algo = Downwind(farm, states, wake_models=[], mbook=self.mbook, verbosity=0)

        results = algo.calc_farm()

        if P_max is not None:
            sname = f"_{type(self).__name__}_set_Pmax"
            self.mbook.turbine_models[sname] = SetFarmVars()
            self.mbook.turbine_models[sname].add_var(FV.MAX_P, P_max)
            models += [sname, "PMask"]

            farm = WindFarm()
            farm.add_turbine(
                Turbine(xy=[0.0, 0.0], turbine_models=models),
                verbosity=0,
            )

            algo = Downwind(farm, states, wake_models=[], mbook=self.mbook, verbosity=0)

            results1 = algo.calc_farm()

            del self.mbook.turbine_models[sname]

        for i, v in enumerate(vars):
            ax = axs[i]
            if ax is None:
                __, ax = plt.subplots(figsize=figsize)

            pargs = {"linewidth": 2.5}
            pargs.update(kwargs)
            ax.plot(ws, results[v][:, 0], label="default", **pargs)
            if P_max is not None:
                pargs = {"linewidth": 1.8}
                pargs.update(pmax_args)
                ax.plot(ws, results1[v][:, 0], label="PMask", **pargs)

            if v == FV.P:
                vv = "Power curve"
            elif v == FV.CT:
                vv = "Thrust curve"
            else:
                vv = v
            t = f"{vv}, {turbine_type}" if titles[i] is None else titles[i]
            ax.set_title(t)

            l = "Wind speed [m/s]" if x_label is None else x_label
            ax.set_xlabel(l)

            if y_labels[i] is None:
                if v == FV.P:
                    l = f"Power [kW]"
                elif v == FV.CT:
                    l = "ct [-]"
                else:
                    l = v
            else:
                l = y_labels[i]
            ax.set_ylabel(l)

            ax.grid()

            if P_max is not None:
                ax.legend()

        return axs



import numpy as np

from foxes.core import TurbineInductionModel
import foxes.variables as FV
import foxes.constants as FC


class SelfSimilar(TurbineInductionModel):
    """
    The self-similar induction wake model
    from Troldborg and Meyer Forsting

    The individual wake effects are superposed linearly,
    without invoking a wake superposition model.

    Notes
    -----
    References:
    [1] Troldborg, Niels, and Alexander Raul Meyer Forsting.
    "A simple model of the wind turbine induction zone derived
    from numerical simulations."
    Wind Energy 20.12 (2017): 2011-2020.
    https://onlinelibrary.wiley.com/doi/full/10.1002/we.2137

    [2] Forsting, Alexander R. Meyer, et al.
    "On the accuracy of predicting wind-farm blockage."
    Renewable Energy (2023).
    https://www.sciencedirect.com/science/article/pii/S0960148123007620

    Attributes
    ----------
    pre_rotor_only: bool
        Calculate only the pre-rotor region
    induction: foxes.core.AxialInductionModel or str
        The induction model

    :group: models.wake_models.induction

    """

    def __init__(self, pre_rotor_only=False, induction="Madsen", gamma=1.1):
        """
        Constructor.

        Parameters
        ----------
        pre_rotor_only: bool
            Calculate only the pre-rotor region
        induction: foxes.core.AxialInductionModel or str
            The induction model
        gamma: float, default=1.1
            The parameter that multiplies Ct in the ct2a calculation

        """
        super().__init__()
        self.induction = induction
        self.pre_rotor_only = pre_rotor_only
        self.gamma = gamma

    def __repr__(self):
        iname = (
            self.induction if isinstance(self.induction, str) else self.induction.name
        )
        return f"{type(self).__name__}(gamma={self.gamma}, induction={iname})"

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            All sub models

        """
        return [self.induction]

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        if isinstance(self.induction, str):
            self.induction = algo.mbook.axial_induction[self.induction]
        super().initialize(algo, verbosity, force)

    def new_wake_deltas(self, algo, mdata, fdata, tdata):
        """
        Creates new empty wake delta arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        wake_deltas: dict
            Key: variable name, value: The zero filled
            wake deltas, shape: (n_states, n_turbines, n_rpoints, ...)

        """
        return {FV.WS: np.zeros_like(tdata[FC.TARGETS][..., 0])}

    def _mu(self, x_R):
        """Helper function: define mu (eqn 11 from [1])"""
        return 1 + (x_R / np.sqrt(1 + x_R**2))

    def _a0(self, ct, x_R):
        """Helper function: define a0 with gamma factor, eqn 8 from [2]"""
        return self.induction.ct2a(self.gamma * ct)

    def _a(self, ct, x_R):
        """Helper function: define axial shape function (eqn 11 from [1])"""
        return self._a0(ct, x_R) * self._mu(x_R)

    def _r_half(self, x_R):
        """Helper function: using eqn 13 from [2]"""
        return np.sqrt(0.587 * (1.32 + x_R**2))

    def _rad_fn(self, x_R, r_R, beta=np.sqrt(2), alpha=8 / 9):
        """Helper function: define radial shape function (eqn 12 from [1])"""
        return (1 / np.cosh(beta * (r_R) / self._r_half(x_R))) ** alpha  # * (x_R < 0)

    def contribute(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        wake_coos,
        wake_deltas,
    ):
        """
        Modifies wake deltas at target points by
        contributions from the specified wake source turbines.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints, ...)

        """
        # get ct
        ct = self.get_data(
            FV.CT,
            FC.STATE_TARGET_TPOINT,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            upcast=True,
            downwind_index=downwind_index,
        )

        # get ws
        ws = self.get_data(
            FV.REWS,
            FC.STATE_TARGET_TPOINT,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            upcast=True,
            downwind_index=downwind_index,
        )

        # get R
        R = 0.5 * self.get_data(
            FV.D,
            FC.STATE_TARGET_TPOINT,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            upcast=False,
            downwind_index=downwind_index,
        )

        # get x, r and R etc. Rounding for safe x < 0 condition below
        x_R = np.round(wake_coos[..., 0] / R, 12)
        r_R = np.linalg.norm(wake_coos[..., 1:3], axis=-1) / R

        # select values
        sp_sel = (ct > 0) & (x_R <= 0)  # upstream
        if np.any(sp_sel):
            # velocity eqn 10 from [1]
            xr = x_R[sp_sel]
            blockage = (
                ws[sp_sel] * self._a(ct[sp_sel], xr) * self._rad_fn(xr, r_R[sp_sel])
            )
            wake_deltas[FV.WS][sp_sel] -= blockage

        # set area behind to mirrored value EXCEPT for area behind turbine
        if not self.pre_rotor_only:
            sp_sel = (ct > 0) & (x_R > 0) & (r_R > 1)
            if np.any(sp_sel):
                # velocity eqn 10 from [1]
                xr = x_R[sp_sel]
                blockage = (
                    ws[sp_sel]
                    * self._a(ct[sp_sel], -xr)
                    * self._rad_fn(-xr, r_R[sp_sel])
                )
                wake_deltas[FV.WS][sp_sel] += blockage

        return wake_deltas
